# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Constants used in the ENUM services."""
from ._enum import StrEnum


class EnumServices(StrEnum):
    """Constants used in the ENUM services."""

    EMAIL_MAILTO = "email mailto"
    EMS_MAILTO = "ems mailto"
    EMS_TEL = "ems tel"
    FAX_TEL = "fax tel"
    FT_FTP = "ft ftp"
    H323 = "H323"
    IFAX_MAILTO = "ifax mailto"
    IM = "im"
    MMS_MAILTO = "mms mailto"
    MMS_TEL = "mms tel"
    PRES = "pres"
    PSTN_SIP = "pstn sip"
    PSTN_TEL = "pstn tel"
    SIP = "SIP"
    SMS_MAILTO = "sms mailto"
    SMS_TEL = "sms tel"
    VOICE_TEL = "voice tel"
    VPIM_LDAP = "VPIM LDAP"
    VPIM_MAILTO = "VPIM MAILTO"
    WEB_HTTP = "web http"
    WEB_HTTPS = "web https"
    XMPP = "xmpp"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the DNS Deployment Role API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class DNSDeploymentRoleType(StrEnum):
    """DNS Deployment Role Type"""

    AD_MASTER = "AD_MASTER"
    FORWARDER = "FORWARDER"
    MASTER = "MASTER"
    MASTER_HIDDEN = "MASTER_HIDDEN"
    NONE = "NONE"
    RECURSION = "RECURSION"
    SLAVE = "SLAVE"
    SLAVE_STEALTH = "SLAVE_STEALTH"
    STUB = "STUB"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Definition of BAM Failover as a target instance."""
from bluecat_libraries.http_client.instance import Instance

__all__ = [
    "FailoverAPIInstance",
]


class FailoverAPIInstance(Instance):
    """
    Definition of BAM Failover as a target instance.

    .. versionadded:: 21.3.1
    """

    def parse_url(self):
        """
        Process the service's URL and construct the value of the base URL of
        the service's API.
        """
        # fmt: off
        self.api_url_base = self.url._replace(
            scheme='https',
            port=self.url.port if self.url.port else '8444',
            path='/'
        )



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# pylint: disable=C0112,C0115,C0116,W0511;
"""Module with exceptions that may occur while working with BlueCat Address Manager and Edge APIs."""


class GeneralError(Exception):
    """
    Base exception to serve as an ancestor for the specific ones.

    :param message: Text about the error that occurred.
    :type message: str
    """

    def __init__(self, message, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = message

    def __str__(self) -> str:
        return self.message


class ClientError(GeneralError):
    """
    Exception that is raised then a client encounters a problem not resulting from
    performing a request or processing its response.

    :param message: Text about the error that occurred.
    :type message: str
    """


class ErrorResponse(GeneralError):
    """
    This exception is raised when an error response is received from a target
    service, e.g., Edge CI API. The HTTP response is available via the `response`
    instance member.

    :param message: Text about the error that occurred.
    :type message: str
    :param response: The HTTP response resulting from a request.
    :type response: requests.Response
    """

    def __init__(self, message, response, *args, **kwargs):
        super().__init__(message, *args, **kwargs)
        self.response = response


class UnexpectedResponse(ErrorResponse):
    """
    This is a specific case of an exception about a response. It is used in the
    cases where the response itself was not an error, but it is not as expected
    based on documentation or contract.

    :param message: Text about the error that occurred.
    :type message: str
    :param response: The HTTP response resulting from a request.
    :type response: requests.Response
    """



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Modules defining values used when working with BlueCat Address Manager."""
from bluecat_libraries.address_manager.constants.access_right_values import AccessRightValues
from bluecat_libraries.address_manager.constants.additional_ip_service_type import (
    AdditionalIPServiceType,
)
from bluecat_libraries.address_manager.constants.dhcp_custom_option_type import DHCPCustomOptionType
from bluecat_libraries.address_manager.constants.dhcp_define_range import DHCPDefineRange
from bluecat_libraries.address_manager.constants.dhcp_deployment_role_type import (
    DHCPDeploymentRoleType,
)
from bluecat_libraries.address_manager.constants.dhcp6_deployment_option_type import (
    DHCP6ClientDeploymentOptionType,
)
from bluecat_libraries.address_manager.constants.dhcp_match_class_criteria import DHCPMatchClass
from bluecat_libraries.address_manager.constants.dhcp_service import (
    DHCPServiceOption,
    DHCPServiceOptionConstant,
)
from bluecat_libraries.address_manager.constants.discovery_type import DiscoveryType
from bluecat_libraries.address_manager.constants.dns_deployment_role_type import (
    DNSDeploymentRoleType,
)
from bluecat_libraries.address_manager.constants.enum_service import EnumServices
from bluecat_libraries.address_manager.constants.ip_group_range_position import IPGroupRangePosition
from bluecat_libraries.address_manager.constants.object_type import ObjectType
from bluecat_libraries.address_manager.constants.option_type import OptionType
from bluecat_libraries.address_manager.constants.response_policy_type import ResponsePolicy
from bluecat_libraries.address_manager.constants.server_capability_profiles import (
    ServerCapabilityProfiles,
)
from bluecat_libraries.address_manager.constants.servers_deployment_status import (
    ServersDeploymentStatus,
    DeploymentTaskStatus,
)
from bluecat_libraries.address_manager.constants.snmp import SNMPVersion
from bluecat_libraries.address_manager.constants.traversal_method import TraversalMethod
from bluecat_libraries.address_manager.constants.user_defined_field_type import UserDefinedFieldType
from bluecat_libraries.address_manager.constants.user_defined_link_type import (
    UserDefinedLinkEntityType,
)
from bluecat_libraries.address_manager.constants.ip_assignment_action_values import (
    IPAssignmentActionValues,
)
from bluecat_libraries.address_manager.constants.ip_address_states import IPAddressState
from bluecat_libraries.address_manager.constants.vendor_profile_option_type import (
    VendorProfileOptionType,
)
from bluecat_libraries.address_manager.constants.zone_template_reapply_mode import (
    ZoneTemplateReapplyMode,
)
from bluecat_libraries.address_manager.constants.defined_probe import (
    DefinedProbeStatus,
    DefinedProbe,
)


__all__ = [
    "AccessRightValues",
    "AdditionalIPServiceType",
    "DefinedProbe",
    "DefinedProbeStatus",
    "DeploymentTaskStatus",
    "DHCP6ClientDeploymentOptionType",
    "DHCPCustomOptionType",
    "DHCPDefineRange",
    "DHCPDeploymentRoleType",
    "DHCPMatchClass",
    "DHCPServiceOption",
    "DHCPServiceOptionConstant",
    "DiscoveryType",
    "DNSDeploymentRoleType",
    "EnumServices",
    "IPAddressState",
    "IPAssignmentActionValues",
    "IPGroupRangePosition",
    "ObjectType",
    "OptionType",
    "ResponsePolicy",
    "ServerCapabilityProfiles",
    "ServersDeploymentStatus",
    "SNMPVersion",
    "TraversalMethod",
    "UserDefinedFieldType",
    "UserDefinedLinkEntityType",
    "VendorProfileOptionType",
    "ZoneTemplateReapplyMode",
]



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values for traversal method used as parameters in APIs."""
from ._enum import StrEnum


class TraversalMethod(StrEnum):
    """Values for traversal method used in APIs that search through hierarchies of objects."""

    BREADTH_FIRST = "BREADTH_FIRST"
    DEPTH_FIRST = "DEPTH_FIRST"
    NO_TRAVERSAL = "NO_TRAVERSAL"



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values for type of user-defined field.

The objects of type 'User-defined field' have a type governing certain characteristics of the field and the values
that the field may hold.
"""
from ._enum import StrEnum


class UserDefinedFieldType(StrEnum):
    """Values for user-defined field parameters."""

    BOOLEAN = "BOOLEAN"
    DATE = "DATE"
    EMAIL = "EMAIL"
    INTEGER = "INTEGER"
    LONG = "LONG"
    TEXT = "TEXT"
    URL = "URL"
    WORKFLOW = "WORKFLOW"



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Structures and functionality related to describing an HTTP client's target instance."""
import urllib3
import logging


__all__ = [
    "Instance",
]

_DEFAULT_SCHEME = "https"


class Instance:
    """
    Definition of a service instance that can be used as a target by a client.

    :param url: URL of the service.
    :type url: str
    :param logger: Logger to be used for logging. This value is currently not
        utilized.
    :type logger: logging.Logger

    :ivar url: URL of the service.
    :vartype url: Url
    :ivar api_url_base: URL to be used as a base for constructing paths to the
        service's endpoints.
    :vartype api_url_base: Url

    .. versionadded:: 20.6.1
    """

    def __init__(self, url: str, logger: logging.Logger = None):
        # pylint: disable=unused-private-member
        self.__logger = logger if logger else logging.getLogger(__name__)
        self.url = urllib3.util.url.parse_url(url)
        self.url = self.url._replace(
            scheme=self.url.scheme if self.url.scheme else _DEFAULT_SCHEME,
            host=self.url.host if self.url.host else "localhost",
            query=None,
            fragment=None,
        )
        self.api_url_base = self.url
        self.parse_url()

    def parse_url(self):
        """
        Process the service's URL and construct the value of the base URL of
        the service's API.
        """



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module for the base client"""

import requests
from requests import codes, Response

from bluecat_libraries.http_client.exceptions import (
    GeneralError,
    ClientError,
    ErrorResponse,
    UnexpectedResponse,
)
from bluecat_libraries.http_client.instance import Instance

__all__ = ["Client"]


class Client:
    """
    An HTTP client with some utilities to make creating service clients easier.
    The class implements the necessary methods to act as a context manager.

    :param target: A definition of a target service.
    :type target: Instance

    .. versionadded:: 20.6.1
    """

    def __init__(self, target: Instance):
        self.session = requests.Session()
        self.target = target
        self.token = None
        self.token_type = None

    def close(self):
        """Release any allocated resources, e.g., an internal session."""
        self.session.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @property
    def is_authenticated(self) -> bool:
        """Determine whether the authentication necessary to communicate with the target service is set."""
        return bool(self.token)

    def _require_auth(self):
        """
        Raise exception if the client does not have the necessary authentication
        set to communicate with the target service.
        """
        if not self.is_authenticated:
            raise ClientError("Use of this method requires authentication, but none is provided.")

    def _url_for(self, name: str) -> str:
        name = name.lstrip("/")
        return self.target.api_url_base.url + name

    def is_error_response(self, response: requests.Response) -> bool:
        """
        Determine whether a response is an error response.

        This method exists to allow derived classes to customize the behaviour.

        :param response: Object representing the HTTP response.
        :type response: requests.Response
        :return: Whether the response conveys an error.
        :rtype: bool
        """
        return not response.ok

    def handle_error_response(self, response: requests.Response):
        """
        Handle a response that's considered to be an error (based on the HTTP
        status code). The standard implementation raises an exception. Derived
        classes may overwrite it to provide custom handling.

        :param response: Object representing the HTTP response.
        :type response: requests.Response
        """
        raise ErrorResponse(
            "Response with HTTP status code {}".format(response.status_code), response
        )

    def http_request(
        self,
        method,
        url,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        timeout=None,
        allow_redirects=True,
        proxies=None,
        hooks=None,
        stream=None,
        verify=None,
        cert=None,
        json=None,
        expected_status_codes=None,
    ):
        """
        Perform an HTTP request based on the provided parameters. It is done
        as part of the internally maintained session, i.e. using the set
        authorization. The majority of the method's parameters correspond to
        their namesakes in :py:meth:`requests.Session.request`. However, this
        class additionally processes the response. It uses
        :py:meth:`is_error_response` and :py:meth:`handle_error_response` to
        detect and handle error responses. That approach allows derived classes
        to customize the behaviour. Additionally, it checks the HTTP status
        code against provided `expected_status_codes` and raises an exception.

        :param method: HTTP method to be used.
        :type method: str
        :param url: URL to be used for the request.
        :type url: str
        :param params: Query parameters to be passed in the request.
        :type params: dict, optional
        :param data: Value to be sent in the body of the request.
        :type data: dict, list[tuple], bytes, or a file-like object, optional
        :param headers: HTTP Headers to be sent with the request.
        :type headers: dict, optional
        :param cookies: Cookies to be sent with the request.
        :type cookies: dict or CookieJar object, optional
        :param files: File-like objects for multipart encoding upload.
        :type files: dict, optional
        :param auth: Object to handle HTTP Authentication.
        :type auth: tuple or callable, optional
        :param timeout: How long to wait to receive data before giving up. If
            given as a tuple, it is used as (connect, read) timeouts.
        :type timeout: float or tuple, optional
        :param allow_redirects: Whether to allow redirects. Defaults to ``True``.
        :type allow_redirects: bool, optional
        :param proxies: Mapping of protocol or protocol and hostname to a URL
            of a proxy to be used.
        :type proxies: dict, optional
        :param hooks: Hooks to be called upon events.
        :type hooks: dict
        :param stream: Whether to immediately read the response content.
            Defaults to ``False``.
        :type stream: bool, optional
        :param verify: Whether to verify the server's TLS certificate. If a
            string is passed, it is treated as a path to a CA bundle to be
            used. Defaults to ``True``.
        :type verify: bool or str
        :param cert: Path to a SSL client certificate file (.pem). If the
            certificate has a key, it can be provides as the second item in a
            tuple, with the first item being the path to the certificate file.
        :type cert: str or tuple, optional
        :param json: Object to be sent as JSON in the body of the request.
        :type json: object, optional
        :param expected_status_codes: HTTP status codes that are acceptable as
            a status code of the HTTP response. If the received code does not
            match the passed values, an :py:exc:`UnexpectedResponse` is raised.
            If left empty, the status code validation is not performed.
        :type expected_status_codes: iterable[int], optional
        :return: Object representing the HTTP response received to the performed
            request.
        :rtype: requests.Response
        """
        try:
            response = self.session.request(
                method,
                url,
                params,
                data,
                headers,
                cookies,
                files,
                auth,
                timeout,
                allow_redirects,
                proxies,
                hooks,
                stream,
                verify,
                cert,
                json,
            )
        except Exception as exc:
            raise GeneralError("Error communicating with target.") from exc

        if self.is_error_response(response):
            self.handle_error_response(response)

        if expected_status_codes and response.status_code not in expected_status_codes:
            raise UnexpectedResponse("Response with unexpected HTTP status code.", response)

        return response

    def http_get(self, url, params=None, expected_status_codes=None, **kwargs) -> Response:
        """
        Perform an HTTP GET request based on the provided parameters. See
        `http_request` for details.

        If `expected_status_codes` is not specified, it defaults to `(200, )`.
        200 is the HTTP status code for successful response `OK`.
        """
        if expected_status_codes is None:
            expected_status_codes = (codes.ok,)
        return self.http_request(
            "GET", url, params, expected_status_codes=expected_status_codes, **kwargs
        )

    def http_post(
        self, url, params=None, data=None, json=None, expected_status_codes=None, **kwargs
    ) -> Response:
        """Perform an HTTP POST request based on the provided parameters. See `http_request` for details."""
        return self.http_request(
            "POST",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_put(
        self, url, params=None, data=None, json=None, expected_status_codes=None, **kwargs
    ) -> Response:
        """Perform an HTTP PUT request based on the provided parameters. See `http_request` for details."""
        return self.http_request(
            "PUT",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_patch(
        self, url, params=None, data=None, json=None, expected_status_codes=None, **kwargs
    ) -> Response:
        """Perform an HTTP PATCH request based on the provided parameters. See `http_request` for details."""
        return self.http_request(
            "PATCH",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_delete(
        self, url, params=None, data=None, json=None, expected_status_codes=None, **kwargs
    ) -> Response:
        """Perform an HTTP DELETE request based on the provided parameters. See `http_request` for details."""
        return self.http_request(
            "DELETE",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module for exposing the API for working with BlueCat Edge."""

from bluecat_libraries.edge.api.client import EdgeClient
from bluecat_libraries.edge.api.exceptions import EdgeErrorResponse

__all__ = [
    "EdgeClient",
    "EdgeErrorResponse",
]



# Copyright 2021-2022 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Constants for versions of BlueCat Address Manager."""

from bluecat_libraries.address_manager.api._version import Version

V9_1_0 = Version("9.1.0")
V9_2_0 = Version("9.2.0")
V9_3_0 = Version("9.3.0")
V9_4_0 = Version("9.4.0")



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in IP Assignment API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class IPAssignmentActionValues(StrEnum):
    """Values for types of actions in BlueCat Address Manager."""

    MAKE_STATIC = "MAKE_STATIC"
    MAKE_RESERVED = "MAKE_RESERVED"
    MAKE_DHCP_RESERVED = "MAKE_DHCP_RESERVED"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the DHCP match class methods in BlueCat Address Manager."""
from ._enum import StrEnum


class DHCPMatchClass(StrEnum):
    """Values for types of DHCP match class."""

    DHCP_CLASS_AGENT_CIRCUIT_ID = "MATCH_AGENT_CIRCUIT_ID"
    DHCP_CLASS_AGENT_REMOTE_ID = "MATCH_AGENT_REMOTE_ID"
    DHCP_CLASS_CLIENT_ID = "MATCH_DHCP_CLIENT_ID"
    DHCP_CLASS_CUSTOM_MATCH = "CUSTOM_MATCH"
    DHCP_CLASS_CUSTOM_MATCH_IF = "CUSTOM_MATCH_IF"
    DHCP_CLASS_HARDWARE = "MATCH_HARDWARE"
    DHCP_CLASS_VENDOR_ID = "MATCH_DHCP_VENDOR_ID"



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Package for exposing a client with autogenerated methods for working with BlueCat Address Manager's API."""

from bluecat_libraries.address_manager.api.rest.auto.client import Client

__all__ = [
    "Client",
]



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Values for sourceEntityTypes and destinationEntityTypes parameters.
The objects of type 'User-Defined Link field' have a type governing certain characteristics of the field and the values
that the field may hold.
"""
from ._enum import StrEnum


class UserDefinedLinkEntityType(StrEnum):
    """
    Values for sourceEntityTypes and destinationEntityTypes parameters.
    """

    DENY_MAC_POOL = "DenyMACPool"
    DEVICE = "Device"
    DHCP4_RANGE = "IP4DHCPRange"
    DHCP6_RANGE = "IP6DHCPRange"
    IP4_ADDRESS = "IP4Addr"
    IP4_BLOCK = "IP4Block"
    IP4_IP_GROUP = "IP4IPGroup"
    IP4_NETWORK = "IP4Network"
    IP4_RANGED = "IP4Ranged"
    IP6_ADDRESS = "IP6Addr"
    IP6_BLOCK = "IP6Block"
    IP6_NETWORK = "IP6Network"
    IP6_RANGED = "IP6Ranged"
    MAC_ADDRESS = "MACAddr"
    MAC_POOL = "MACPool"
    SERVER = "Server"
    SERVER_GROUP = "ServerGroup"
    VIEW = "View"
    ZONE = "Zone"



# Copyright 2021-2022 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Models representing all Address Manager object types supported in the API."""
import copy
import json

from bluecat_libraries.address_manager.api.serialization import (
    deserialize_joined_key_value_pairs,
    serialize_joined_key_value_pairs,
    serialize_joined_values,
    serialize_possible_list,
    deserialize_possible_list,
)


class APIEntity(dict):
    """
    Model for the BAM API object type APIEntity.

    :key id: The entity's ID. Value type is int.
    :key name: The entity's name. Value type is str | None.
    :key type: The entity's type. Must be a valid BAM object type.
    :key properties: (Optional) Additional properties on the entity. Value must be dict[str, str].
    """

    @staticmethod
    def to_raw_model(data: dict) -> dict:
        """
        :param data: APIEntity object or dict equivalent.
        :type data: dict[str, Any]
        :return: Dict that, once JSON-serialized, can be passed to BAM endpoints.
        :rtype: dict[str, str]
        """
        data = copy.deepcopy(data)
        data["properties"] = serialize_joined_key_value_pairs(data.get("properties"))
        return data

    @staticmethod
    def from_raw_model(data: dict) -> "Optional[APIEntity]":
        """
        :param data: Dict obtained by JSON-deserializing an APIEntity returned by a BAM endpoint.
        :type data: dict[str, str]
        :return: Entity object, or None if input's ``id`` is 0.
        :rtype: APIEntity | None
        """
        data = copy.deepcopy(data)
        if data["id"] == 0:
            return None
        data["properties"] = (
            deserialize_joined_key_value_pairs(data["properties"]) if data.get("properties") else {}
        )
        return APIEntity(data)


class APIAccessRight(dict):
    """
    Model for the BAM API object type APIAccessRight.

    :key entityId: ID of the object to which the access right applies. Value type is int. Must be greater than 0.
    :key userId: The access right's owner's ID. Value type is int. Must be greater than 0.
    :key value: Value must be "HIDE", "VIEW", "ADD", "CHANGE", or "FULL".
    :key overrides: (Optional) Override access rights of ``entityId``'s descendants (by default,
        they inherit ``entityId``'s access right). Value type is dict[str, str]. Keys are object
        types to be overriden; values are access right values.
    :key properties: (Optional) Additional properties on the access right. Value type is dict[str, str].
    """

    @staticmethod
    def to_raw_model(data: dict) -> dict:
        """
        :param data: APIAccessRight object or dict equivalent.
        :type data: dict[str, Any]
        :return: Dict that, once JSON-serialized, can be passed to BAM endpoints.
        :rtype: dict[str, str]
        """
        data = copy.deepcopy(data)
        data["overrides"] = serialize_joined_key_value_pairs(data.get("overrides"))
        data["properties"] = serialize_joined_key_value_pairs(data.get("properties"))
        return data

    @staticmethod
    def from_raw_model(data: dict) -> "APIAccessRight":
        """
        :param data: Dict obtained by JSON-deserializing an APIAccessRight returned by a BAM endpoint.
        :type data: dict[str, str]
        :return: Access right object.
        :rtype: APIAccessRight
        """
        data = copy.deepcopy(data)
        data["overrides"] = (
            deserialize_joined_key_value_pairs(data["overrides"]) if data.get("overrides") else {}
        )
        data["properties"] = (
            deserialize_joined_key_value_pairs(data["properties"]) if data.get("properties") else {}
        )
        return APIAccessRight(data)


class APIDeploymentRole(dict):
    """
    Model for the BAM API object type APIDeploymentRole.

    :key id: The deployment role's ID. Value type is int.
    :key type: Value must be "NONE", "MASTER", "MASTER_HIDDEN", "SLAVE", "SLAVE_STEALTH", "FORWARDER", "STUB", "RECURSION", "PEER", or "AD MASTER".
    :key service: Value must be "DNS", "DHCP", or "TFTP".
    :key entityId: The deployed entity's ID. Value type is int. Must be greater than 0.
    :key serviceInterfaceId: ID of the server interface being deployed into. Value type is int. Must be greater than 0.
    :key properties: (Optional) Additional properties on the deployment role. Value type is dict[str, str].
    """

    @staticmethod
    def to_raw_model(data: dict) -> dict:
        """
        :param data: APIDeploymentRole object or dict equivalent.
        :type data: dict[str, Any]
        :return: Dict that, once JSON-serialized, can be passed to BAM endpoints.
        :rtype: dict[str, str]
        """
        data = copy.deepcopy(data)
        data["properties"] = serialize_joined_key_value_pairs(data.get("properties"))
        return data

    @staticmethod
    def from_raw_model(data: dict) -> "Optional[APIDeploymentRole]":
        """
        :param data: Dict obtained by JSON-deserializing an APIDeploymentRole returned by a BAM endpoint.
        :type data: dict[str, str]
        :return: Deployment role object, or None if input's ``id`` is 0.
        :rtype: APIDeploymentRole | None
        """
        data = copy.deepcopy(data)
        if data.get("id") == 0:
            return None
        data["properties"] = (
            deserialize_joined_key_value_pairs(data["properties"]) if data.get("properties") else {}
        )
        return APIDeploymentRole(data)


class APIDeploymentOption(dict):
    """
    Model for the BAM API object type APIDeploymentOption.

    :key id: The deployment option's ID. Value type is int.
    :key type: The deployment option's type. Must be a valid BAM option type.
    :key name: The deployment option's name. Value type is str.
    :key value: Field values of the option. Value type is list[str].
    :key properties: (Optional) Additional properties on the deployment option. Value type is dict[str, str].
    """

    @staticmethod
    def to_raw_model(data: dict) -> dict:
        """
        :param data: APIDeploymentOption object or dict equivalent.
        :type data: dict[str, Any]
        :return: Dict that, once JSON-serialized, can be passed to BAM endpoints.
        :rtype: dict[str, str]
        """
        data = copy.deepcopy(data)
        data["value"] = serialize_possible_list(data.get("value", ""))
        data["properties"] = serialize_joined_key_value_pairs(data.get("properties"))
        return data

    @staticmethod
    def from_raw_model(data: dict) -> "Optional[APIDeploymentOption]":
        """
        :param data: Dict obtained by JSON-deserializing an APIDeploymentOption returned by a BAM endpoint.
        :type data: dict[str, str]
        :return: Deployment role object, or None if input's ``id`` is 0.
        :rtype: APIDeploymentOption | None
        """
        data = copy.deepcopy(data)
        if data.get("id") == 0:
            return None
        data["value"] = deserialize_possible_list(data.get("value", ""))
        data["properties"] = (
            deserialize_joined_key_value_pairs(data["properties"]) if data.get("properties") else {}
        )
        return APIDeploymentOption(data)


class APIUserDefinedField(dict):
    """
    Model for the BAM API object type APIUserDefinedField.

    :key name: The UDF's unique name. Value type is str.
    :key displayName: The UDF's display name. Value type is str.
    :key type: The UDF's type. Must be a valid BAM UDF type.
    :key defaultValue: The UDF's default value. Value type is str.
    :key required: If true, users must enter data in the field. Value type is bool.
    :key hideFromSearch: If true, the UDF is hidden from search. Value type is bool.
    :key validatorProperties: (Optional) The UDF's validation properties. Value type is dict[str, str].
    :key predefinedValues: (Optional) The UDF's preset values. Value type is list[str].
    :key properties: (Optional) Additional properties on the UDF. Value type is dict[str, str].
    """

    @staticmethod
    def to_raw_model(data: dict) -> dict:
        """
        :param data: APIUserDefinedField object or dict equivalent.
        :type data: dict[str, Any]
        :return: Dict that, once JSON-serialized, can be passed to BAM endpoints.
        :rtype: dict[str, str]
        """
        data = copy.deepcopy(data)
        data["predefinedValues"] = serialize_joined_values(
            data.get("predefinedValues"), item_sep="|"
        )
        data["validatorProperties"] = serialize_joined_key_value_pairs(
            data.get("validatorProperties"), item_sep=","
        )  # object types that can take multiple properties, separate each property with a “,” comma
        data["properties"] = serialize_joined_key_value_pairs(data.get("properties"))
        return data

    @staticmethod
    def from_raw_model(data: dict) -> "APIUserDefinedField":
        """
        :param data: Dict obtained by JSON-deserializing an APIUserDefinedField returned by a BAM endpoint.
        :type data: dict[str, str]
        :return: UDF object.
        :rtype: APIUserDefinedField
        """
        data = copy.deepcopy(data)

        if data.get("predefinedValues"):
            data["predefinedValues"] = list(filter(None, data.get("predefinedValues").split("|")))
        else:
            data["predefinedValues"] = []

        if data.get("validatorProperties"):
            data["validatorProperties"] = deserialize_joined_key_value_pairs(
                data.get("validatorProperties"), item_sep=","
            )
        else:
            data["validatorProperties"] = {}

        data["properties"] = (
            deserialize_joined_key_value_pairs(data["properties"]) if data.get("properties") else {}
        )
        return APIUserDefinedField(data)


class UDLDefinition(dict):
    """
    Model for the structure describing User-Defined Link definitions used by
    Address Manager's API.

    .. note:: Starting from Address Manager v9.4.0, the option **description** is supported.

        * The maximum length of this field is 512 characters.
        * If the **description** field is not specified, it will be null.
        * On upgrading from Address Manager v9.3.0, the **description** for the existing links will
          be populated as **null**.

    :key linkType: The UDL's unique name. Value type is str. Cannot be a reserved link type name
        and cannot start with "BCN\\_".
    :key displayName: The UDL's name as displayed in BAM. Value type is str.
    :key sourceEntityTypes: The UDL's source entity types. Value type is list[str].
    :key destinationEntityTypes: The UDL's destination entity types. Value type is list[str].
    :key description: (Optional) The description for the user-defined link.
    """

    # NOTE: The use of '\\_' in the above docstring is intentional. The goal is 2 level escaping:
    # 1) '\\' translates into '\' in the Python string
    # 2) '\_' make reStructuredText not treat 'BCN_' as an internal hyperlink target.

    @staticmethod
    def to_raw_model(data: dict) -> str:
        """
        :param data: UDLDefinition object or dict equivalent.
        :type data: dict[str, Any]
        :return: JSON-encoded string that can be passed to BAM endpoints.
        :rtype: str
        """
        return json.dumps(data)

    @staticmethod
    def from_raw_model(data: dict) -> "UDLDefinition":
        """
        :param data: Dict obtained by JSON-deserializing a UDL returned by a BAM endpoint.
        :type data: dict[str, str]
        :return: UDL definition object.
        :rtype: UDLDefinition
        """
        return UDLDefinition(data)


class UDLRelationship(dict):
    """
    Model for the structure describing User-Defined Link relationships used by
    Address Manager's API.

    .. note:: Starting from Address Manager v9.4.0, the option **description** and
        **destinationEntityIds** are supported.

    :key linkType: The UDL's link type. Value type is str.
    :key sourceEntityId: The UDL's source entity ID. Value type is int.
    :key description: (optional) The description for the user-defined link.
    :key destinationEntityId: (Optional) The UDL's destination entity ID. Value type is int.
    :key destinationEntityIds: (Optional) The list of UDL's destination entity ID, comma-separated
        inside square brackets.
    """

    @staticmethod
    def to_raw_model(data: dict) -> str:
        """
        :param data: UDLRelationship object or dict equivalent.
        :type data: dict[str, Any]
        :return: JSON-encoded string that can be passed to BAM endpoints.
        :rtype: str
        """
        return json.dumps(data)

    @staticmethod
    def from_raw_model(data: dict) -> "UDLRelationship":
        """
        :param data: Dict obtained by JSON-deserializing the result of UDLRelationship.to_raw_model(<something>).
        :type data: dict[str, str]
        :return: UDL relationship object.
        :rtype: UDLRelationship
        """
        return UDLRelationship(data)


class RetentionSettings(dict):
    """
    Model for BAM history retention settings.

    :key admin: (Optional) The number of days of administrative history to keep in the database. Value type is int.
    :key sessionEvent: (Optional) The number of days of session event history to keep in the database. Value type is int.
    :key ddns: (Optional) The number of days of DDNS history to keep in the database. Value type is int.
    :key dhcp: (Optional) The number of days of DHCP history to keep in the database. Value type is int.

    .. note::

        * The value for sessionEvent must be greater than or equal to the value of
          each of the other types.
        * The input value for the retention periods (in days) must be greater than or equal to one.
        * Setting the value to -1 is equivalent to Retain Indefinitely in the BAM database.
        * Setting the DDNS and DHCP retention setting to 0 is equivalent to Do Not Retain,
          and these records no longer write to the database.
          So, if a user has enabled the audit data export feature, they will get no records written to their audit data.
    """

    @staticmethod
    def to_raw_model(data: dict) -> dict:
        """
        :param data: RetentionSettings object or dict equivalent.
        :type data: dict[str, Any]
        :return: Dict that, once JSON-serialized, can be passed to BAM endpoints.
        :rtype: dict
        """
        data = copy.deepcopy(data)
        update_admin = data.get("admin") is not None
        update_session_event = data.get("sessionEvent") is not None
        update_ddns = data.get("ddns") is not None
        update_dhcp = data.get("dhcp") is not None
        return dict(
            admin=data.get("admin"),
            updateAdmin=update_admin,
            sessionEvent=data.get("sessionEvent"),
            updateSessionEvent=update_session_event,
            ddns=data.get("ddns"),
            updateDdns=update_ddns,
            dhcp=data.get("dhcp"),
            updateDhcp=update_dhcp,
        )

    @staticmethod
    def from_raw_model(data: str) -> "RetentionSettings":
        """
        :param data: A value in the format returned by BAM method "updateRetentionSettings"
            that holds the ordered settings for: admin, sessionEvent, ddns, and dhcp.
        :type data: str
        :return: Retention settings object.
        :rtype: RetentionSettings
        """
        admin, session_event, ddns, dhcp = list(map(int, data.split(",")))
        return RetentionSettings(
            admin=admin,
            sessionEvent=session_event,
            ddns=ddns,
            dhcp=dhcp,
        )


class ResponsePolicySearchResult(dict):
    """Model for the BAM API object type ResponsePolicySearchResult.

    :key configId: ID of the parent configuration in which the response policy item is configured. Value type is int.
    :key parentIds: IDs of parent response policy or response policy zone objects. Value type is list[int].
        If policy item is associated with a Response Policy, it is the Response Policy object ID.
        If policy item is associated with BlueCat Security feed data, it is the RP Zone object ID.
    :key name: The response policy item's name. Value type is str.
    :key category: The name of the BlueCat security feed category associated with the policy item. Value type is str | None.
    :key policyType: The response policy's type. Value type is str.
    """

    @staticmethod
    def from_raw_model(data: dict) -> "ResponsePolicySearchResult":
        """
        :param data: Dict obtained by JSON-deserializing a ResponsePolicySearchResult returned by a BAM endpoint.
        :type data: dict[str, str]
        :return: Response policy search result object.
        :rtype: ResponsePolicySearchResult
        """
        data = copy.deepcopy(data)
        data["parentIds"] = list(map(int, data.get("parentIds").split(",")))
        return ResponsePolicySearchResult(data)


class APIData(dict):
    """
    Model for the BAM API object type APIData.

    :key name: The name of the probe to collect data. Value type is str.
    :key properties: Additional properties on the probe. Value must be list.
    """

    @staticmethod
    def from_raw_model(data: dict) -> "APIData":
        """
        :param data: Dict obtained by JSON-deserializing an APIData returned by a BAM endpoint.
        :type data: dict[str, str]
        :return: API Data object.
        :rtype: APIData
        """
        data = copy.deepcopy(data)
        data["properties"] = json.loads(data["properties"])
        return APIData(data)



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Wrapper for standard list REST results to make them mimic `suds` list results.
"""

from bluecat_libraries.address_manager.api.rest.provisional.rest_dict import RESTDict


class RESTEntityArray(list):
    """
    Wrapper for standard list REST results to make them mimic `suds` list results.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for index, entity in enumerate(self):
            if isinstance(entity, dict):
                self[index] = RESTDict(entity)

    @property
    def item(self):
        """
        Property provided for compatibility with `suds` list result.
        Points at the instance itself.
        """
        return self



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Module for the GraphQL client
For internal BlueCat use only!
"""
from __future__ import annotations

import json
from string import Template
from typing import Any, Optional

import requests
from bluecat_libraries.address_manager._graphql.exceptions import GraphQLException
from bluecat_libraries.http_client.exceptions import ClientError


class Client:
    """
    BAM GraphQL client
    """

    def __init__(self, url: str, *, verify: bool | str = True) -> None:
        """
        ``verify`` has the same meaning as arguments of the same name commonly found in
        ``requests`` functions.
        """
        self.url = url

        self._verify = verify
        self._session: Optional[requests.Session] = None

    @property
    def is_authenticated(self) -> bool:
        """Whether or not this client is currently authenticated with a session."""
        return bool(self._session)

    def invalidate(self) -> None:
        """
        Log out of session and release any allocated session resources. If the client was not
        authenticated, this is a no-op.
        """
        if not self.is_authenticated:
            return
        # TDB: make logout request
        self._session.close()
        self._session = None

    def __enter__(self) -> Client:
        return self

    def __exit__(self, *args) -> None:
        self.invalidate()

    def get_tokens(self) -> dict[str, str]:
        """
        Return authentication tokens for the client's current session. These can be used to
        authenticate another ``Client`` object without having to make an authentication request.
        """
        if not self.is_authenticated:
            raise ClientError("Client is not authenticated")
        return {
            "jsession_id": self._session.cookies["JSESSIONID"],
            "nonce": self._session.headers["X-NONCE"],
        }

    def set_tokens(self, *, jsession_id: str, nonce: str) -> Client:
        """
        If the client is not authenticated, create a new session that's authenticated by the given
        tokens. Otherwise, replace tokens on the existing session. Return the client afterwards.
        """
        if not self._session:
            self._session = requests.Session()
        self._session.cookies["JSESSIONID"] = jsession_id
        self._session.headers["X-NONCE"] = nonce
        return self

    @staticmethod
    def __raise_if_errors(response: requests.Response) -> dict[str, Any]:
        """
        Check for errors in the GraphQL response and raise an exception if any exist. Otherwise,
        return the response body.
        """
        if not response.ok:
            raise GraphQLException(
                "Response with HTTP status code " + str(response.status_code), response, None
            )
        body = response.json()
        if "errors" in body:
            message = json.dumps([e["message"] for e in body["errors"]], indent=2)
            raise GraphQLException(message, response, body["errors"])
        return body

    def authenticate(self, username: str, password: str) -> Client:
        """Authenticate with a session, then return the client."""
        graph_query = Template(
            """
            mutation {
                basicAuth(
                    input: {
                        username: "$username",
                        password: "$password",
                        startSession: true
                    }
                ) {
                    session { nonce }
                }
            }
            """
        ).substitute(username=username, password=password)
        response = requests.post(self.url, json={"query": graph_query}, verify=self._verify)
        body = Client.__raise_if_errors(response)
        return self.set_tokens(
            jsession_id=response.cookies["JSESSIONID"],
            nonce=body["data"]["basicAuth"]["session"]["nonce"],
        )

    def execute(self, query: str, variables: Optional[dict] = None) -> dict:
        """
        Execute the given query (or mutation). If the query requires authentication, this client
        must first be authenticated with a session using the ``authenticate()`` method.

        :param query: GraphQL API query
        :type query: str
        :param variables: Arguments for the GraphQL API query
        :type variables: dict
        :return: A dictionary containing the GraphQL query response.
        :rtype: dict
        """
        payload = {"query": query}
        if variables:
            payload["variables"] = variables
        response = (self._session or requests).post(self.url, json=payload, verify=self._verify)
        return Client.__raise_if_errors(response)



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the DHCP service options API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class DHCPServiceOption(StrEnum):
    """Constants used in the DHCP service option methods."""

    ALLOW_DHCP_CLASS_MEMBERS = "allow-dhcp-class-members"
    ALLOW_MAC_POOL = "allow-mac-pool"
    ALWAYS_BROADCAST = "always-broadcast"
    ALWAYS_REPLY_RFC1048 = "always-reply-rfc1048"
    APPLY_MAC_AUTHENTICATION_POLICY = "apply-mac-authentication-policy"
    CLIENT_UPDATES = "client-updates"
    CONFLICT_DETECTION = "conflict-detection"
    DDNS_DOMAINNAME = "ddns-domainname"
    DDNS_DUAL_STACK_MIXED_MODE = "ddns-dual-stack-mixed-mode"
    DDNS_GUARD_ID_MUST_MATCH = "ddns-guard-id-must-match"
    DDNS_HOSTNAME = "ddns-hostname"
    DDNS_REV_DOMAINNAME = "ddns-rev-domainname"
    DDNS_TTL = "ddns-ttl"
    DDNS_UPDATES = "ddns-updates"
    DDNS_UPDATE_STYLE = "ddns-update-style"
    DEFAULT_LEASE_TIME = "default-lease-time"
    DENY_DHCP_CLASS_MEMBERS = "deny-dhcp-class-members"
    DENY_DHCP_CLIENTS = "deny-dhcp-clients"
    DENY_MAC_POOL = "deny-mac-pool"
    DDNS_OTHER_GUARD_IS_DYNAMIC = "ddns-other-guard-is-dynamic"
    DENY_UNKNOWN_MAC_ADDRESSES = "deny-unknown-mac-addresses"
    DHCP_CLASS_LEASE_LIMIT = "dhcp-class-lease-limit"
    DO_REVERSE_UPDATES = "do-reverse-updates"
    DYNAMIC_BOOTP_LEASE_LENGTH = "dynamic-bootp-lease-length"
    FILENAME = "filename"
    GET_LEASE_HOSTNAMES = "get-lease-hostnames"
    LOAD_BALANCE_OVERRIDE = "load-balance-override"
    LOAD_BALANCE_SPLIT = "load-balance-split"
    MAX_LEASE_TIME = "max-lease-time"
    MAX_RESPONSE_DELAY = "max-response-delay"
    MAX_UNACKED_UPDATES = "max-unacked-updates"
    MCLT = "mclt"
    MIN_LEASE_TIME = "min-lease-time"
    MIN_SECS = "min-secs"
    NEXT_SERVER = "next-server"
    ONE_LEASE_PER_CLIENT = "one-lease-per-client"
    PING_CHECK = "ping-check"
    SERVER_IDENTIFIER = "server-identifier"
    SITE_OPTION_SPACE = "site-option-space"
    STASH_AGENT_OPTIONS = "stash-agent-options"
    UPDATE_CONFLICT_DETECTION = "update-conflict-detection"
    UPDATE_OPTIMIZATION = "update-optimization"
    UPDATE_STATIC_LEASES = "update-static-leases"
    USE_LEASE_ADDR_FOR_DEFAULT_ROUTE = "use-lease-addr-for-default-route"


class DHCPServiceOptionConstant(StrEnum):
    """
    Values used in DHCP service options. Depending on the type of deployment option being added,
    the format of the value input might differ.
    """

    DDNS_HOSTNAME_POSITION_APPEND = "append"
    DDNS_HOSTNAME_POSITION_PREPEND = "prepend"
    DDNS_HOSTNAME_TYPE_DUID = "duid"
    DDNS_HOSTNAME_TYPE_FIXED = "fixed"
    DDNS_HOSTNAME_TYPE_IP = "ip"
    DDNS_HOSTNAME_TYPE_MAC = "mac"
    DDNS_UPDATE_STYLE_INTERIM = "interim"
    DDNS_UPDATE_STYLE_STANDARD = "standard"



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Module for the client to work with BlueCat Edge's API.
"""
# pylint: disable=too-many-lines
from requests import Response, codes
import urllib
from bluecat_libraries.edge.api.instance import EdgeInstance
from bluecat_libraries.edge.api.exceptions import EdgeErrorResponse
from bluecat_libraries.http_client import Client, ClientError, UnexpectedResponse

__all__ = [
    "EdgeClient",
]


class EdgeClient(Client):
    """`Client` uses the `requests` library to communicate with the REST endpoints of DNS Edge."""

    def __init__(self, url):
        super().__init__(EdgeInstance(url))

    def handle_error_response(self, response: Response):
        """
        Handle a response that's considered to be an error. If the response
        matches an Edge error, an `EdgeErrorResponse` is raised. Otherwise,
        the handling is delegated to the base class' implementation.
        """
        if response.headers["Content-Type"].lower() == "application/json":
            try:
                data = response.json()
                error_message = data["brief"]
                error_code = data["code"]
            except Exception as exc:
                raise UnexpectedResponse(
                    "DNS Edge API returned an error that cannot be processed.", response
                ) from exc
            raise EdgeErrorResponse(error_message, response, error_code)
        super().handle_error_response(response)

    def clear_token(self):
        """Clear token from client and session header"""
        self.token = None
        self.token_type = None
        if "Authorization" in self.session.headers:
            del self.session.headers["Authorization"]

    # region Authentication
    def authenticate(self, client_id: str, client_secret: str, refresh_token: bool = False):
        """
        Log in to the DNS Edge CI using the access key set to retrieve a token and bearer type to populate the
        authorization header.

        :param client_id: Edge client ID.
        :type client_id: str, required
        :param client_secret: Edge secret access key.
        :type client_secret: str, required
        :param refresh_token: Retrieve a new token even if logged in.
        :type refresh_token: bool, optional

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)

        .. versionadded:: 20.6.1
        """
        if self.is_authenticated and not refresh_token:
            raise ClientError("Client is already authenticated.")

        payload = {
            "grantType": "ClientCredentials",
            "clientCredentials": {"clientId": client_id, "clientSecret": client_secret},
        }

        response = self.http_post(
            self._url_for("/v1/api/authentication/token"),
            json=payload,
            expected_status_codes=(codes.ok,),
        )

        data = response.json()
        self.token = data["accessToken"]
        self.token_type = data["tokenType"]
        self.session.headers["Authorization"] = self.token_type + " " + self.token

    def get_v1_api_authentication_token_describe(self):
        """
        Return information about the user whose credentials were used to obtain
        the authorization token.

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    current_user = client.get_v1_api_authentication_token_describe()

                print(current_user)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/authentication/token/describe"))
        return response.json()

    # endregion Authentication

    # region DNS Query Stats
    def get_v1_api_customer_dnsquerystats_count(
        self, start: int, downsample: int, end: int = None
    ) -> list:
        """
        Return the total count of logged DNS queries per policy action type in a Customer Instance for any time
        interval within the last 24 hours.

        :param start: The start of the time range, in milliseconds since the Unix Epoch.
        :type start: int, required
        :param downsample: The sample period in the returned dataset, in minutes.
        :type downsample: int, required
        :param end: The end of the time range, in milliseconds since the Unix Epoch.
        :type end: int, optional
        :return: DNS query count results including columns: time, totalQueries, totalAllowed, totalMonitored,
         totalRedirected, totalNonMatched
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                start = 0         # Start of time range
                downsample = 10   # Sample period of returned dataset

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    stats = client.get_v1_api_customer_dnsquerystats_count(start=start, downsample=downsample)

                for stat in stats:
                    print(stat)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v1/api/customer/dnsQueryStats/count"),
            params={"start": start, "end": end, "downsample": downsample},
        )
        return response.json()

    def get_v1_api_customer_dnsquerystats_topdomains(
        self, start: int, count: int, end: int = None
    ) -> list:
        """
        Return the specified number of domains most frequently queried per Customer Instance for any time interval
        within the last 24 hours.

        :param start: The start of the time range, in milliseconds since the Unix Epoch.
        :type start: int, required
        :param count: The number of domains to return.
        :type count: int, required
        :param end: The end of the time range, in milliseconds since the Unix Epoch.
        :type end: int, optional
        :return: The most frequently queried domains info: domainName, hitcount
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                start = 0         # Start of time range
                count = 10        # Number of results to return

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_customer_dnsquerystats_topdomains(start=start, count=count)

                for domain in result:
                    print(domain)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v1/api/customer/dnsQueryStats/topDomains"),
            params={"start": start, "end": end, "count": count},
        )
        return response.json()

    def get_v1_api_customer_dnsquerystats_uniqueip(self, start: int, end: int = None) -> dict:
        """
        Return the total number of unique client IPs that issued DNS queries in a Customer Instance for any time
        interval within the last 24 hours.

        :param start: The start of the time range, in milliseconds since the Unix Epoch.
        :type start: int, required
        :param end: The end of the time range, in milliseconds since the Unix Epoch. Defaults to `None`.
        :type end: int, optional
        :return: Total number of unique client IPs that issued DNS queries in a Customer Instance with key "count".
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                start = 0         # Start of time range

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    stats = client.get_v1_api_customer_dnsquerystats_uniqueip(start=start)

                print(stats["count"])

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v1/api/customer/dnsQueryStats/uniqueIp"),
            params={"start": start, "end": end},
        )
        return response.json()

    # endregion

    # region Service Point Fleet Management
    def get_v1_api_servicepoints(self) -> list:
        """
        Return information about all currently registered service points.

        :return: Currently registered service points information including: id, name, siteId, lastSync, connectionState,
            connectionStateLastChanged, ipAddresses, loopbackIps, version, updateInitiatedTimestamp, updateStatus
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_servicepoints()

                for sp in result:
                    print(sp)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/servicePoints"))
        return response.json()

    # endregion

    # region audit logs
    def get_v1_api_audit_logs(
        self,
        offset: str = None,
        limit: int = None,
        start_time: str = None,
        end_time: str = None,
        user_name: str = None,
    ) -> list:
        """
        Return the audit log records of API requests and responses.

        :param offset: The number of records to skip, from the beginning of the log.
        :type offset: str, optional
        :param limit: The maximum number of records to retrieve.
        :type limit: int, optional
        :param start_time: The start time of the period to return logs for, in ISO_8601 format.
        :type start_time: str, optional
        :param end_time: The end time of the period to return logs for, in ISO_8601 format.
        :type end_time: str, optional
        :param user_name: To filter log records for a particular user, specify a user name.
        :type user_name: str, optional
        :return: Audit log records of API requests and responses.
        :rtype: list

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_audit_logs()

                for log in result:
                    print(log)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v1/api/audit/logs"),
            params={
                "offset": offset,
                "limit": limit,
                "startTime": start_time,
                "endTime": end_time,
                "username": user_name,
            },
        )
        return response.json()

    # endregion

    # region Site Administration
    def get_v3_api_sites(
        self,
        name: str = None,
        name_contains: str = None,
        fields: str = None,
        desired_result_count: int = None,
        namespace_id: str = None,
        overrides_forwarders: bool = None,
    ) -> dict:
        """
        Return information about all the sites.

        :param name: Filter by the exact name of a site
        :type name: str, optional
        :param name_contains:  Filter sites by part of a site name. The name and name_contains parameters
            are mutually exclusive and shouldn't be used at the same time.
        :type name_contains: str, optional
        :param fields: Fields to include in the response.
            Example: ?fields=id,name will return a response that contains only the sites' IDs and names.
        :type fields: str, optional
        :param desired_result_count: The number of sites to include in the response.
        :type desired_result_count: int, optional
        :param namespace_id: Filter sites by a specific namespace ID.
        :type namespace_id: str, optional
        :param overrides_forwarders: Determine whether to include sites that have overriding forwarders
            for the specified namespace. Default is false. This parameter is only effective when
            used with the namespaceId parameter.
        :type overrides_forwarders: bool, optional
        :return: Sites information including: id, name, location, settings, registeredServicePointCount,
            version, updateInitiatedTimestamp, updateStatus, blockedServicePointIds
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v3_api_sites()

                for sp in result:
                    print(sp)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        params = dict(
            name=name,
            nameContains=name_contains,
            fields=fields,
            desiredResultCount=desired_result_count,
            namespaceId=namespace_id,
            overridesForwarders=overrides_forwarders,
        )
        response = self.http_get(self._url_for("/v3/api/sites"), params=params)
        return response.json()

    def get_v3_api_sites_by_id(self, site_id: str) -> dict:
        """
        Return information of the site that matches the specified site ID.

        :param site_id: The ID of the site.
        :type site_id: str, required
        :return: Information of the site including: id, name, location, settings, registeredServicePointCount,
            updateStatus, updateInitiatedTime, version, blockedServicePointIds
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_id = '3f76fe19-900d-47ff-b00c-c3cab6f40903'         # Site ID

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v3_api_sites_by_id(site_id=site_id)

                print(result)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v3/api/sites/" + site_id))
        return response.json()

    def patch_v3_api_sites(self, payload: list):
        """
        Update the service point version of specified sites.

        .. note:: You must disable the automatic service point updates to use this API.

        :param payload: A list of the sites whose service points should be updated. It is
            possible to exclude certain service points from the operation. Each
            item in the list represents a site and options for it.
        :type payload: list[dict]

        Example:

            .. code-block:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = [
                    {
                        "id": "33709b0b-d429-45f1-8c82-51b2942a9133",
                        "version": "3.0.5",
                        "blockedServicePointIds": [
                            "feac4ad6-adb6-4c8d-924e-d314d29ec96d",
                            "91cbda2f-ac60-4a32-bc9f-02161e532b2d"]
                    },
                    {
                        "id": "aaa0830d-4639-4210-9158-0ac6395e2396",
                        "version": "3.0.6"
                    }
                ]

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.patch_v3_api_sites(payload)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_patch(
            self._url_for("/v3/api/sites"), json=payload, expected_status_codes=(codes.no_content,)
        )

    def post_v3_api_sites(self, payload: dict) -> str:
        """
        Create a new site.

        :param payload: Payload used to create a site.
        :type payload: dict, required
        :return: The ID of the created site.
        :rtype: str

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = {
                    "name": "<site-name>",
                    "location": {
                    "address": "<address>"
                    }
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v3_api_sites(payload=payload)

                print(result)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_post(
            self._url_for("/v3/api/sites"), json=payload, expected_status_codes=(codes.created,)
        )
        site_id = response.headers["Location"].rsplit("/", 1)[-1]
        return site_id

    def put_v3_api_sites(self, site_id: str, payload: dict):
        """
        Update an existing site. Note that all the parameters are updated with this
        action, whether you specify the values or not. If you leave a parameter
        out, its value will be overwritten with no data.

        :param site_id: The ID of the site.
        :type site_id: str
        :param payload: Site parameters and their new values.
        :type payload: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_id = "28c1b312-55e0-4dc4-8253-de7ae37db25d"
                payload = {
                    "name": "the new site name",
                    "location": {
                         "address": "123 Main St., Toronto",
                         "lng": "-79.3756671",
                         "lat": "43.6421529"
                     }
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.put_v3_api_sites(site_id, payload)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        self.http_put(
            self._url_for("/v3/api/sites/") + site_id,
            json=payload,
            expected_status_codes=(codes.no_content,),
        )

    def delete_v3_api_sites(self, site_id: str):
        """
        Delete the site specified by the site ID.

        :param site_id: The ID of the site to delete.
        :type site_id: str, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_id = "<site_id>"

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.delete_v3_api_sites(site_id=site_id)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        self.http_delete(
            self._url_for("/v3/api/sites/" + site_id), expected_status_codes=(codes.no_content,)
        )

    def post_v3_api_sites_clearcache(self, site_id: str):
        """
        Clear the cache of a site.

        :param site_id: The ID of the site.
        :type site_id: str

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_id = "28c1b312-55e0-4dc4-8253-de7ae37db25d"

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.post_v3_api_sites_clearcache(site_id)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        self.http_post(
            self._url_for("/v3/api/sites/" + site_id + "/clearCache"),
            expected_status_codes=(codes.accepted,),
        )

    # endregion

    # region Site Groups Administration
    def get_v1_api_customer_sitesandsitegroups_search(
        self, name_contains: str = None, desired_result_count: int = None
    ) -> list:
        """
        Return a list of the sites that match the specified search string.

        :param name_contains: Search for site and site group names which contain the specified string.
        :type name_contains: str, required
        :param desired_result_count: The maximum number of records to return.
        :type desired_result_count: int, required
        :return: A list of the sites that match the specified search string.
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                name_contains = "test"         # search key word
                desired_result_count = 10      # max number of records to return

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_customer_sitesandsitegroups_search(name_contains=name_contains,
                                                                              desired_result_count=desired_result_count)

                for s in result:
                    print(s)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v1/api/customer/sitesAndSiteGroups/search"),
            params={"nameContains": name_contains, "desiredResultCount": desired_result_count},
        )
        return response.json()

    def post_v1_api_customer_sitegroups_create(self, payload: dict) -> dict:
        """
        Create a new site group with the sites specified by site ID.

        :param payload: Payload used to create a site group.
        :type payload: dict, required
        :return: The details of the site group created.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient
                payload = {
                    "name":"name",
                    "description":"description",
                    "siteIds":["id1","id2"]
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v1_api_customer_sitegroups_create(payload=payload)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_post(self._url_for("/v1/api/customer/siteGroups/create"), json=payload)
        return response.json()

    def post_v1_api_customer_sitegroups_delete(self, site_group_id: int):
        """
        Delete the site group specified by site group ID.

        :param site_group_id: The ID of the site group.
        :type site_group_id: int, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_group_id = 2

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v1_api_customer_sitegroups_delete(site_group_id=site_group_id)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_post(
            self._url_for("/v1/api/customer/siteGroups/" + str(site_group_id) + "/delete")
        )

    def post_v1_api_customer_sitegroups_update(self, site_group_id: int, payload: dict):
        """
        Update the site group specified by site group ID with the new name, description, or sites.

        :param site_group_id: The ID of the site group.
        :type site_group_id: int, required
        :param payload: Payload used to update the site group.
        :type payload: dict, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_group_id = 2
                payload = {
                    "name":"name",
                    "description":"description",
                    "siteIds":["id1","id2"]
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v1_api_customer_sitegroups_update(site_group_id=site_group_id, payload=payload)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_post(
            self._url_for("/v1/api/customer/siteGroups/" + str(site_group_id) + "/update"),
            json=payload,
        )

    def get_v1_api_customer_sitegroups(self) -> list:
        """
        Gets a list of the site groups.

        :return: A list of all site groups.
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_customer_sitegroups()

                for sp in result:
                    print(sp)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/customer/siteGroups"))
        return response.json()

    # endregion

    # region API Access Key Set
    def get_v1_api_apikeys(self) -> list:
        """
        Return the API access key sets for the current user.

        :return: API access key sets for the current user.
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    results = client.get_v1_api_apikeys()

                for result in results:
                    print(result)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/apiKeys"))
        return response.json()

    def post_v1_api_apikeys(self) -> dict:
        """
        Create an API access key set for the current user.

        :return: Dictionary containing clientId and clientSecret that is created.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    api_key = client.post_v1_api_apikeys()

                print(api_key)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_post(self._url_for("/v1/api/apiKeys"))
        return response.json()

    def delete_v1_api_apikey_by_id(self, client_id):
        """
        Delete the API access key set with the specified client ID for the current user.

        :param client_id: The client id that needs to be deleted.
        :type client_id: str, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.delete_v1_api_apikey_by_id(<client_id>)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        self.http_delete(self._url_for("/v1/api/apiKeys/{id}".format(id=client_id)))

    def delete_v1_api_apikey(self, email_address):
        """
        Delete all the API access key sets for the specified email address.

        :param email_address: Delete all the API access key sets associated with this email address
        :type email_address: str, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.delete_v1_api_apikey(<email_address>)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        self.http_delete(
            self._url_for("/v1/api/apiKeys"),
            params=dict(email_address=email_address),
            expected_status_codes=(codes.no_content,),
        )

    # endregion

    # region Service Point Versions
    def get_v1_api_spversions(self) -> list:
        """
        Return a list of available service point versions, listed from the most recently released version to the oldest
        supported version.

        :return: A list of available service point versions.
        :rtype: list[str]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_spversions()

                for ver in result:
                    print(ver)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/spVersions"))
        return response.json()

    # endregion

    # region Domain Lists
    def get_v2_domainlists(self, name_contains: str = None) -> list:
        """
        Return a list of domain lists (that match the specified search string).

        :param name_contains: Search for domains which contain the specified string.
        :type name_contains: str, optional
        :return: A list of domain lists (that match the specified search string).
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v2_domainlists()

                for r in result:
                    print(r)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v2/domainLists"), params={"nameContains": name_contains}
        )
        return response.json()

    def post_v2_domainlists(self, payload: dict) -> str:
        """
        Create a domain list.

        :param payload: Payload used to create a domain list.
        :type payload: dict, required
        :return: The ID of a new domain list.
        :rtype: str

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = {
                    "name": "<domain-list-name>",
                    "description": "<domain-list-description>",
                    "sourceType": "user"
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v2_domainlists(payload=payload)

                print(result)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_post(
            self._url_for("/v2/domainLists"), json=payload, expected_status_codes=(codes.created,)
        )
        domain_list_id = response.headers["Location"].rsplit("/", 1)[-1]
        return domain_list_id

    def delete_v2_domainlists(self, domain_list_id: str):
        """
        Delete the domain list specified by the domain list ID.

        :param domain_list_id: Payload used to create a domain list.
        :type domain_list_id: str, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                domain_list_id = "<domain_list_id>"

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.delete_v2_domainlists(domain_list_id=domain_list_id)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        self.http_delete(
            self._url_for("/v2/domainLists/" + domain_list_id),
            expected_status_codes=(codes.no_content,),
        )

    def get_v2_domainlists_by_id(self, domain_list_id: str) -> dict:
        """
        Return a specific domain list.

        :param domain_list_id: The ID of the domain list to return.
        :type domain_list_id: str, required
        :return: Information of the domain list (that matches the given id) including: id, name, description,
            sourceType, domainCount, etc.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                domain_list_id = "<domain_list_id>"

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v2_domainlists_by_id(domain_list_id=domain_list_id)

                print(result)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v2/domainLists/" + domain_list_id))
        return response.json()

    def put_v2_domainlists(self, domain_list_id: str, domain_names: list):
        """
        Replace all domains within a domain list by supplying a list of domain names.

        :param domain_list_id: The ID of the domain list to update.
        :type domain_list_id: str, required
        :param domain_names: The names of domains to replace existing with.
        :type domain_names: list, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                domain_list_id = "<domain_list_id>"
                domain_names = ["name-001", "name-002"]

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.put_v2_domainlists(domain_list_id=domain_list_id, domain_names=domain_names)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        data = "\n".join(domain_names)
        self.http_put(
            self._url_for("/v2/domainLists/" + domain_list_id),
            data=data,
            expected_status_codes=(codes.no_content,),
        )

    def patch_v2_domainlists(
        self, domain_list_id: str, domain_names_add: list = None, domain_names_remove: list = None
    ):
        """
        Update the content of an existing domain list by adding or deleting domain names.

        :param domain_list_id: The ID of the domain list to update.
        :type domain_list_id: str, required
        :param domain_names_add: The names of domains to add to the domain list.
        :type domain_names_add: list, optional
        :param domain_names_remove: The names of domains to remove from the domain list.
        :type domain_names_remove: list, optional

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                domain_list_id = "<domain_list_id>"
                domain_names_add = ["name-001", "name-002"]
                domain_names_remove = ["name-003", "name-004"]

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.patch_v2_domainlists(domain_list_id=domain_list_id, domain_names_add=domain_names_add,
                                                domain_names_remove=domain_names_remove)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        if domain_names_add is None:
            domain_names_add = []
        if domain_names_remove is None:
            domain_names_remove = []
        data = "\n".join(domain_names_add)
        if domain_names_remove:
            data = [data, "-", *domain_names_remove]
            data = "\n".join(data)
        self.http_patch(
            self._url_for("/v2/domainLists/" + domain_list_id),
            data=data,
            expected_status_codes=(codes.no_content,),
        )

    def get_v1_api_list_dns_by_id(self, domain_list_id: str) -> list:
        """
        Return a list of domain lists.

        :param domain_list_id: The ID of the domain list.
        :type domain_list_id: str, required
        :return: Domain names of the given domain list.
        :rtype: list

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                domain_list_id = '3f76fe19-900d-47ff-b00c-c3cab6f40903'         # domain list ID

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.v1_api_list_dns_by_id(domain_list_id=domain_list_id)

                for r in result:
                    print(r)

        .. versionadded:: 20.6.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/list/dns/" + domain_list_id))
        return response.text.strip("\n").split("\n")

    def get_v1_api_list_dns(self) -> list:
        """
        Return a list of domain lists.

        :return: A list of domain lists
        :rtype: list

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient


                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_list_dns()

                for r in result:
                    print(r)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/list/dns"))
        return response.json()

    def post_v1_api_list_dns(self, payload: dict) -> dict:
        """
        Create a domain list.

        :param payload: Payload used to create a domain list.
        :type payload: dict, required
        :return: The created domain list.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = {
                    "name": "<domain-list-name>",
                    "description": "<domain-list-description>",
                    "sourceType": "user"
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v1_api_list_dns(payload=payload)

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_post(self._url_for("/v1/api/list/dns"), json=payload)
        return response.json()

    def post_v1_api_list_dns_sourceconfiguration(self, domain_list_id: str, payload: dict):
        """
        Create a domain list sourced and dynamically updated from a local feed.

        :param domain_list_id: The ID of the domain list.
        :type domain_list_id: str, required
        :param payload: Source configuration payload.
        :type payload: dict, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = {
                            "hostName":"source.host.name.org",
                            "privateKey":"privateKey",
                            "filePath":"/location/location",
                            "port":22,
                            "transferType":"rsyncSSH",
                            "user":"userName",
                            "publicHostKey":"publicKey",
                            "syncRate":5
                        }
                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v1_api_list_dns_sourceconfiguration(<domain_list_id>, payload)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_post(
            self._url_for(f"/v1/api/list/dns/{domain_list_id}/sourceConfiguration"), json=payload
        )

    def post_v1_api_list_dns_attachfile(self, domain_list_id: str, file_handle):
        """
        Upload a CSV file containing a list of domain names to an existing DNS list.

        :param domain_list_id: The ID of the domain list.
        :type domain_list_id: str, required
        :param file_handle: File handle that contains the domains.
        :type file_handle: file, required
        :return: A dictionary containing a number of registered domains.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    with open('file.csv', 'r') as f:
                        client.authenticate(<client_id>, <client_secret>)
                        result = client.post_v1_api_list_dns_attachfile(<domain_list_id>, f)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_post(
            self._url_for(f"/v1/api/list/dns/{domain_list_id}/attachfile"),
            files={"file": ("file.txt", file_handle, "text/plain")},
        )
        return response.json()

    def post_v1_api_list_dns_delete(self, domain_list_id: str):
        """
        Delete the domain list specified by the domain list ID.

        :param domain_list_id: The ID of the domain list.
        :type domain_list_id: str, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v1_api_list_dns_delete(<domain_list_id>)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_post(self._url_for(f"/v1/api/list/dns/{domain_list_id}/delete"))

    def get_v1_api_list_dns_search(self, name_contains: str, desired_result_count: int):
        """
        Return a list of domain lists with names containing the search fragment, stopping when the desired
        result count reaches.

        :param name_contains: Filter based on the name.
        :type name_contains: str, required
        :param desired_result_count: The maximum number of returned results.
        :type desired_result_count: int, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_list_dns_search(<name_contains>, 10)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v1/api/list/dns/search"),
            params=dict(nameContains=name_contains, desiredResultCount=desired_result_count),
        )
        return response.json()

    # endregion

    # region Policy Management
    def get_v5_api_policies(
        self, site_id: str = None, site_group_id: str = None, domain_list_id: str = None
    ) -> list:
        """
        Get a list of all policies, or policies associated with a specified site, site group, or domain list.

        :param site_id: Specify a site by site ID to search for policies associated with a site.
        :type site_id: str, optional
        :param site_group_id: Specify a site group by site group ID to search for policies associated with a site group.
        :type site_group_id: str, optional
        :param domain_list_id: Specify a domain list by domain list ID to search for policies associated with a domain list.
        :type domain_list_id: str, optional

        :return: A list of policies
        :rtype: list

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v5_api_policies()

                for r in result:
                    print(r)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v5/api/policies"),
            params=dict(siteId=site_id, siteGroupId=site_group_id, domainListId=domain_list_id),
        )
        return response.json()

    def post_v5_api_policies(self, payload: dict) -> str:
        """
        Create a new policy.

        :param payload: Payload used to create a policy.
        :type payload: dict, required

        :return: The ID of the created policy
        :rtype: str

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.post_v5_api_policies(<payload>)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_post(
            self._url_for("/v5/api/policies"), json=payload, expected_status_codes=(codes.created,)
        )
        policy_id = response.headers["Location"].split("/")[-1]
        return policy_id

    def get_v5_api_policies_by_id(self, policy_id: str) -> dict:
        """
        Search for a policy by policy ID.

        :param policy_id: Specify a policy by ID to return.
        :type policy_id: str, required

        :return: Policy for the given policy ID.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v5_api_policies_by_id(<policy_id>)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v5/api/policies/" + policy_id))
        return response.json()

    def delete_v5_api_policies(self, policy_id: str):
        """
        Delete the policy specified by the policy ID.

        :param policy_id: The ID of the policy to delete.
        :type policy_id: str, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                policy_id = "<policy_id>"

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.delete_v5_api_policies(policy_id)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_delete(
            self._url_for("/v5/api/policies/" + policy_id),
            expected_status_codes=(codes.ok, codes.no_content),
        )

    def put_v5_api_policies(self, policy_id: str, payload: dict):
        """
        Update the policy specified by the policy ID. Note that all the fields are updated with this action,
        whether you specify any values or not. If you leave a field out, it's value will be overwritten with no data.

        :param policy_id: The ID of the policy to update.
        :type policy_id: str, required
        :param payload: Payload used to update the policy.
        :type payload: dict, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                policy_id = "<policy_id>"

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.put_v5_api_policies(policy_id, <payload>)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_put(
            self._url_for("/v5/api/policies/" + policy_id),
            json=payload,
            expected_status_codes=(codes.ok, codes.no_content),
        )

    # endregion

    # region Namespaces
    def get_v1_api_namespaces(self, site_id: str = "") -> list:
        """
        Return a list of namespaces, or namespaces associated with the site specified by site ID, if specified.

        :param site_id: The ID of the site.
        :type site_id: str, optional
        :return: The namespaces (all or those associated with the given site).
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_namespaces()

                for ns in result:
                    print(ns)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/namespaces"), params={"siteId": site_id})
        return response.json()

    def get_v1_api_namespaces_by_id(self, namespace_id: str) -> dict:
        """
        Return information of the namespace that matches the specified namespace ID.

        :param namespace_id: The ID of the namespace.
        :type namespace_id: str, required
        :return: Information of the namespace including: id, name, description, forwarders, matchLists, exceptionLists,
            isDefault, associatedSiteSettings, ttl, umbrellaIntegrationId, latency
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                namespace_id = <namespace_id>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_namespaces_by_id(namespace_id=namespace_id)

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/namespaces/" + namespace_id))
        return response.json()

    def post_v1_api_namespaces(self, payload: dict) -> str:
        """
        Create a namespace.

        :param payload: Payload used to create a namespace.
        :type payload: dict, required
        :return: The ID of the created namespace.
        :rtype: str

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = {
                   "name": "namespace",
                   "description": "this is a namespace",
                   "forwarders": ["8.8.8.8", "2.2.2.2"],
                   "matchLists": ["domainListId1", "domainListId2"],
                   "exceptionLists": ["domainListId1", "domainListId2"],
                   "umbrellaIntegrationId": "<id of umbrella integration>",
                   "ttl": 60
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v1_api_namespaces(payload=payload)

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_post(
            self._url_for("/v1/api/namespaces"),
            json=payload,
            expected_status_codes=(codes.created,),
        )
        namespace_id = response.headers["Location"].rsplit("/", 1)[-1]
        return namespace_id

    def put_v1_api_namespaces(self, namespace_id: str, payload: dict):
        """
        Update the namespace specified by the namespace ID. Note that all the fields are updated with this action, whether
        you specify values or not. If you leave a field out, its value will be overwritten with no data.

        :param namespace_id: The ID of the namespace to update.
        :type namespace_id: str, required
        :param payload: Payload used to update a namespace.
        :type payload: dict, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                namespace_id = <namespace_id>
                payload = {
                   "name": "namespace",
                   "description": "this is a namespace",
                   "forwarders": ["8.8.8.8", "2.2.2.2"],
                   "matchLists": ["domainListId1", "domainListId2"],
                   "exceptionLists": ["domainListId1", "domainListId2"],
                   "umbrellaIntegrationId": "<id of umbrella integration>",
                   "ttl": 60
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.put_v1_api_namespaces(namespace_id=namespace_id, payload=payload)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_put(
            self._url_for("/v1/api/namespaces/" + namespace_id),
            json=payload,
            expected_status_codes=(codes.no_content,),
        )

    def patch_v1_api_namespaces(self, namespaces: list):
        """
        Set (up to three) namespaces as defaults.

        :param namespaces: List of namespace names and flags.
        :type namespaces: list[dict], required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                namespaces = [
                    {
                        "id": "<id of the namespace>",
                        "isDefault": "true" or "false"
                    },
                    {
                        "id": "<id of the other namespace>",
                        "isDefault": "true" or "false"
                    }
                ]

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.patch_v1_api_namespaces(namespaces=namespaces)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_patch(
            self._url_for("/v1/api/namespaces"),
            json=namespaces,
            expected_status_codes=(codes.no_content,),
        )

    def delete_v1_api_namespaces(self, namespace_id: str):
        """
        Delete the namespace specified by the namespace ID.

        :param namespace_id: The ID of the namespace to delete.
        :type namespace_id: str, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                namespace_id = <namespace_id>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.delete_v1_api_namespaces(namespace_id=namespace_id)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_delete(
            self._url_for("/v1/api/namespaces/" + namespace_id),
            expected_status_codes=(codes.no_content,),
        )

    # endregion

    # region Terms of Service
    def get_v1_api_tos(self):
        """
        Return the name, user email address, and timestamp when the current user accepted their Terms of Service.

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.get_v1_api_tos()

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/tos"))
        return response.json()

    # endregion

    # region DNS Query Logs
    def get_v3_api_dnsquerylogs(self, filters: dict = None) -> list:
        """
        Return a list of all (or optionally filtered) logged DNS queries.

        :param filters: The filters to narrow down the search result. The following filters (optional) can be used:

            * site_id (string or list) - The ID of the site to query for logs.
            * batch_size (integer) - The maximum number of records to return. Default size is 400. The maximum
              configurable size is 10000.
            * key (string) - The record ID of the DNS query log record to receive older or newer records.
            * order (string) - Specify whether to retrieve results in descending (default) or ascending order.
            * start_time (string) - Filter results for a specific time period, include the start time,
              in milliseconds since the Unix Epoch.
            * end_time (string) - Filter results for a specific time period, include the end time,
              in milliseconds since the Unix Epoch.
            * has_matched_policy (bool) - If true, only the DNS queries matching policies will display.
            * source_ip (string) - Filter results for the specified source IP address.
            * query_type (string) - The query record type to search for.
            * query_name (string) - The domain name to search for queries.
            * policy_action (string) - If one or more policy actions are provided, then only the policies matching the
              actions are returned. Valid actions are allow, block, redirect, and monitor.
            * policy_name (string) - If a policy name is provided, then only the policies matching the name are returned.
            * policy_id (string): If a policy ID is provided, then only the policies matching the ID are returned.
            * threat_type (string) - Return queries that match the specified threat type.
              Valid threat types are DGA and DNS_TUNNELING.
            * threat_indicator (string) - Return queries that match the specified threat indicator.
              Valid threat indicators are ENTROPY, UNIQUE_CHARACTER, EXCEEDING_LENGTH, UNCOMMON_QUERY_TYPE,
              VOLUMETRIC_TUNNELING, SUSPECT_DNS, and SUSPECT_TLD.
            * response_code (string) - The response code of the DNS query.
            * protocol (string) - The protocol of the DNS query (usually UDP or TCP).
            * namespace_id (string) - The ID of namespaces a DNS query was queried against.

        :type filters: dict, optional
        :return: The DNS queries (all or filtered).
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                filters = {
                    "site_id": [<site_id_1>, <site_id_2>],
                    "source_ip": "172.17.149.40"
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v3_api_dnsquerylogs(filters=filters)

                for q in result:
                    print(q)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        if filters is None:
            filters = {}
        params = {}
        param_names = {
            "site_id": "siteId",
            "batch_size": "batchSize",
            "key": "key",
            "order": "order",
            "start_time": "startTime",
            "end_time": "endTime",
            "has_matched_policy": "hasMatchedPolicy",
            "source_ip": "sourceIp",
            "query_type": "queryType",
            "query_name": "queryName",
            "policy_action": "policyAction",
            "policy_name": "policyName",
            "policy_id": "policyId",
            "threat_type": "threatType",
            "threat_indicator": "threatIndicator",
            "response_code": "responseCode",
            "protocol": "protocol",
            "namespace_id": "namespaceId",
        }

        for k in filters:
            if k in param_names and filters[k]:
                if isinstance(filters[k], list):
                    val = list(map(str, filters[k]))
                elif isinstance(filters[k], bool):
                    val = "true" if filters[k] else "false"
                else:
                    val = str(filters[k])
                params[param_names[k]] = val

        response = self.http_get(self._url_for("/v3/api/dnsQueryLogs"), params=params)
        return response.json()

    def get_v3_api_dnsquerylogs_by_id(self, query_id: str) -> dict:
        """
        Return the details for a specific query.

        :param query_id: The ID of the query.
        :type query_id: str, required
        :return: Information of the DNS query including: time, source, siteId, query, queryType, response, id,
            actionTaken, matchedPolicies, authority, queryProtocol, threats, queriedNamespaces, latency
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                query_id = <query_id>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v3_api_dnsquerylogs_by_id(query_id=query_id)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v3/api/dnsQueryLogs/" + query_id))
        return response.json()

    def get_v2_api_customer_dnsquerylog_count(self, site_name: str) -> dict:
        """
        Returns the total count of logged DNS queries within 30 days for the specified site name.

        :param site_name: Site name to search for the count of DNS query logs
        :type site_name: str, required
        :return: Count of logged DNS queries within 30 days for the specified site name.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_name = <site_name>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v2_api_customer_dnsquerylog_count(site_name=site_name)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(
            self._url_for("/v2/api/customer/dnsQueryLog/count"), params=dict(siteName=site_name)
        )
        return response.json()

    # endregion

    # region Settings
    def get_v1_api_settings(self):
        """
        Retrieve the automatic service point update settings.

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    api_settings = client.get_v1_api_settings()

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/settings"))
        return response.json()

    def put_v1_api_settings(self, payload: dict):
        """
        Update the automatic service point update settings.

        :param payload: API parameters and their values
        :type payload: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = {
                    "spAutomaticUpdatesEnabled": "true" or "false"
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.put_v1_api_settings(payload)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_put(
            self._url_for("/v1/api/settings"),
            json=payload,
            expected_status_codes=(codes.no_content,),
        )

    # endregion

    # region System Lists
    def get_v1_api_list_system(self, list_name: str) -> list:
        """
        Return the specified system-maintained domain lists, including all of the domains in the list,
        and the last time each listed domain incurred DNS traffic identifying it as suspected tunneling.

        :param list_name: The name of the list
        :type list_name: str, required
        :return: The list of all the domains and the last time each listed domain incurred DNS traffic.
        :rtype: list

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                query_id = <query_id>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_list_system('tunneling')

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/list/system/" + list_name))
        return response.json()

    def get_v1_api_list_system_content(self, list_name: str) -> str:
        """
        Return the contents of the specified system-maintained domain lists in JSON or CSV format. Leave out the
        Accept parameter to retrieve a CSV list instead of JSON.
        For each domain, the last time the domain incurred DNS traffic identifying it as a suspect for tunneling,
        and the time when that domain clears from the list according to the system-level set TTL,
        are returned.

        :param list_name: The name of the list
        :type list_name: str, required
        :return: The contents of the specified system-maintained domain lists.
        :rtype: list

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                query_id = <query_id>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_list_system_content('tunneling')

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/list/system/content/" + list_name))
        return response.text

    def get_v1_api_list_system_information(self, list_name: str) -> dict:
        """
        Return the meta information for the specified system-maintained domain list.

        :param list_name: The name of the list
        :type list_name: str, required
        :return: The system-maintained domain list meta information.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                query_id = <query_id>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_list_system_information('tunneling')

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/list/system/information/" + list_name))
        return response.json()

    # endregion

    def get_v2_api_spimage_config(self, site_id: str) -> str:
        """
        Download the base64 encoded service point configuration information, including:

         - the endpoint to connect to for registration
         - the certificate that the service point uses to authenticate to its endpoint
         - the service point manager endpoint

        :param site_id: The id of the site
        :type site_id: str, required
        :return: The base64 encoded service point configuration information.
        :rtype: str

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_id = <site_id>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v2_api_spimage_config(site_id=site_id)

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v2/api/spImage/" + site_id + "/config"))
        return response.text

    def post_v2_api_spimage_generate(self, payload: dict):
        """
        Retrieve an OVA image of the specified site.

        :param payload: Payload (that contains site id) used to generate the OVA image.
        :type payload: dict, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = {
                    "siteId": <site_id>
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v2_api_spimage_generate(payload=payload)

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_post(self._url_for("/v2/api/spImage/generate"), json=payload)

    def get_v2_api_spimage_status(self, site_id: str) -> dict:
        """
        Retrieve the status of the image generation.

        :param site_id: The id of the site
        :type site_id: str, required
        :return: The status of the image generation along with other information.
        :rtype: dict

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                site_id = <site_id>

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v2_api_spimage_status(site_id=site_id)

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v2/api/spImage/" + site_id + "/status"))
        return response.json()

    def post_v1_api_list_dns_update(self, domain_list_id: str, payload: dict):
        """
        Update a domain list name, description, or source type.

        :param domain_list_id: The ID of the domain list to update.
        :type domain_list_id: str, required
        :param payload: Payload used to update the domain list.
        :type payload: dict, required

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                domain_list_id = "<domain_list_id>"
                payload = {
                   "name": "new domain list name",
                   "description": "new domain list description",
                   "sourceType": "dynamic"
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.post_v1_api_list_dns_update(domain_list_id, payload)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        self.http_post(
            self._url_for("/v1/api/list/dns/" + domain_list_id + "/update"), json=payload
        )

    # region Users
    def get_v1_api_users(self) -> list:
        """
        Return the list of users.

        :return: A list of all users.
        :rtype: list[dict]

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.get_v1_api_users()

                for user in result:
                    print(user)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_get(self._url_for("/v1/api/users"))
        return response.json()

    def post_v1_api_users(self, payload: dict) -> str:
        """
        Create a new user. An email is sent to the user with a temporary
        password and instructions on how to log in and set a new password.

        :param payload: Values for the user fields.
        :type payload: dict, required
        :return: The ID of the created user.
        :rtype: str

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                payload = {
                    "email": "email address",
                    "name": "first and last name",
                    "role": "SYSADMIN | ADMIN | ANALYST",
                    "status": "active | inactive"
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    result = client.post_v1_api_users(payload=payload)

                print(result)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        response = self.http_post(
            self._url_for("/v1/api/users"), json=payload, expected_status_codes=(codes.created,)
        )
        user_id = response.headers["Location"].rsplit("/", 1)[-1]
        # The user ID is known to contain a pipe character, which is URL-encoded
        # as "%7C" and needs to be unquoted.
        user_id = urllib.parse.unquote(user_id)
        return user_id

    def put_v1_api_users(self, user_id: str, payload: dict) -> dict:
        """
        Update the email address, name, role, and status of the user specified
        by the user ID.

        :param user_id: ID of the user to update.
        :type user_id: str
        :param payload: Values for the user fields.
        :type payload: dict
        :return: The updated user.
        :rtype: dict

        .. note:: You can't update the email address of other users. You can
            only update your email address.

        Example:

            .. code:: python

                from bluecat_libraries.edge.api import EdgeClient

                user_id = <user_id>
                payload = {
                    "email":"new email address",
                    "name":"new name",
                    "role":"ADMIN | ANALYST",
                    "status":"active | inactive"
                }

                with EdgeClient(<edge_ci_url>) as client:
                    client.authenticate(<client_id>, <client_secret>)
                    client.put_v1_api_users(user_id=user_id, payload=payload)

        .. versionadded:: 20.12.1
        """
        self._require_auth()
        # The user ID is known to contain a pipe character, which needs to be
        # URL-encoded as "%7C" before being used in the URL.
        user_id = urllib.parse.quote(user_id)
        response = self.http_put(
            self._url_for("/v1/api/users/" + user_id),
            json=payload,
            expected_status_codes=(codes.ok,),
        )
        return response.json()

    # endregion Users



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# pylint: disable=C0114,C0115,C0116;
"""Values related to the HTTP requests to and responses from BlueCat DNS Edge."""

import enum


class ErrorResponseCode(enum.Enum):
    """Values found as `code` in an error response from BlueCat DNS Edge."""

    # fmt: off
    BAD_REQUEST = 'BAD_REQUEST'                                # Request is invalid
    CLIENT_CREDENTIALS_INVALID = 'CLIENT_CREDENTIALS_INVALID'  # Client credentials are invalid
    ERROR_INVALID_ARGUMENT = 'ERROR_INVALID_ARGUMENT'          # Invalid request.
    INVALID_TOKENS_REQUEST = 'INVALID_TOKENS_REQUEST'          # Invalid token request body
    UNAUTHORIZED_ACTION = 'UNAUTHORIZED_ACTION'                # You are not authorized to perform this action
    USER_ALREADY_EXISTS = 'USER_ALREADY_EXISTS'                # User already exists.
    SITE_DOES_NOT_EXIST = 'SITE_DOES_NOT_EXIST'                # Site does not exist



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in methods related to zone template in BlueCat Address Manager."""
from ._enum import StrEnum


class ZoneTemplateReapplyMode(StrEnum):
    """Constants used in methods with zone template."""

    UPDATE = "templateReapplyModeUpdate"
    IGNORE = "templateReapplyModeIgnore"
    OVERWRITE = "templateReapplyModeOverwrite"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Modules for working with BlueCat Address Manager's Failover API."""

from bluecat_libraries.http_client import Client as HTTPClient
from bluecat_libraries.address_manager.failover_api.instance import FailoverAPIInstance


class Client(HTTPClient):
    """
    A client for calling the Failover API of Address Manager.

    The Failover API is a service configuration in Address Manager that can
    enable an externally automated failover when the primary Address Manager
    fails.

    :param url: The base URL of BAM's Failover service.
    :type url: str
    :param cert: Path to an SSL client certificate. The first item being the path to the certificate file
        and the second item is the key.
    :type cert: tuple
    :param verify: Whether to verify the server's TLS certificate. If a
        string is passed, it is treated as a path to a CA bundle to
        use. Defaults to ``True``.
    :type verify: bool or str

    .. versionadded:: 21.5.1
    """

    def __init__(self, url: str, cert, verify=True):
        super().__init__(target=FailoverAPIInstance(url=url))
        self.session.cert = cert
        self.session.headers = {"Accept": "application/json"}

        self.session.verify = verify

    def get_v1_health(self, timeout: int = None) -> dict:
        """
        Return the health status of Address Manager.
        This API can be called on primary, secondary, or standalone Address Manager servers.

        :param timeout: The timeout of this request in milliseconds.
        :type timeout: int
        :return: Returns the health status of Address Manager.
        :rtype: dict

        Example:

            .. code-block:: python

                from bluecat_libraries.address_manager.failover_api import Client

                client_cert = ('client_certificate.crt', 'key.pem')
                server_cert = 'server_certificate.crt'
                with Client(url='https://<bam_address>:8444', cert=client_cert, verify=server_cert) as client:
                    data = client.get_v1_health()
                print(data)

        .. versionadded:: 21.5.1
        """
        params = None
        if timeout:
            params = dict(timeout=timeout)
        response = self.http_get(
            url=self._url_for("/bam/v1/health"),
            params=params,
            cert=self.session.cert,
            verify=self.session.verify,
        )
        return response.json()

    def put_v1_promote(self) -> str:
        """
        Promote a secondary Address Manager server in the cluster to primary.

        :return: Returns the health status of Address Manager if successful.
        :rtype: str

        Example:

            .. code-block:: python

                from bluecat_libraries.address_manager.failover_api import Client

                client_cert = ('client_certificate.crt', 'key.pem')
                server_cert = 'server_certificate.crt'
                with Client(url='https://<bam_address>:8444', cert=client_cert, verify=server_cert) as client:
                    result = client.put_v1_promote()
                print(result)

        .. versionadded:: 21.5.1
        """
        response = self.http_put(
            url=self._url_for("/bam/v1/promote"), cert=self.session.cert, verify=self.session.verify
        )
        return response.json()

    def put_v1_managed_servers_takeover(self) -> list:
        """
        Notify all DNC/DHCP Servers controlled by the old primary Address Manager server
        of the new primary Address Manager server.

        :return: list

        Example:

            .. code-block:: python

                from bluecat_libraries.address_manager.failover_api import Client

                client_cert = ('client_certificate.crt', 'key.pem')
                server_cert = 'server_certificate.crt'
                with Client(url='https://<bam_address>:8444', cert=client_cert, verify=server_cert) as client:
                    data = client.put_v1_managed_servers_takeover()
                print(data)

        .. versionadded:: 21.5.1
        """
        response = self.http_put(
            url=self._url_for("/bam/v1/managed-servers/takeover"),
            cert=self.session.cert,
            verify=self.session.verify,
        )
        return response.json()

    def put_v1_managed_servers_takeover_by_id(self, server_id: int):
        """
        Notify a single DNC/DHCP Server controlled by the old primary Address Manager server
        of the new primary Address Manager server.

        :param server_id: The ID of the DNS/DHCP Server.
        :type server_id: int
        :return: object

        Example:

            .. code-block:: python

                from bluecat_libraries.address_manager.failover_api import Client

                client_cert = ('client_certificate.crt', 'key.pem')
                server_cert = 'server_certificate.crt'
                server_id = 123
                with Client(url='https://<bam_address>:8444', cert=client_cert, verify=server_cert) as client:
                    data = client.put_v1_managed_servers_takeover_by_id(server_id)
                print(data)

        .. versionadded:: 21.5.1
        """
        response = self.http_put(
            url=self._url_for(f"/bam/v1/managed-servers/{server_id}/takeover"),
            cert=self.session.cert,
            verify=self.session.verify,
        )
        return response.json()



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module for the client to work with BlueCat Service Point API."""

from bluecat_libraries.http_client.client import Client
from bluecat_libraries.http_client.exceptions import UnexpectedResponse
from bluecat_libraries.service_point.api.instance import ServicePointInstance


class ServicePointClient(Client):
    """`Client` uses the `requests` library to communicate with Service Point."""

    def __init__(self, url):
        super().__init__(ServicePointInstance(url))

    def get_v1_status_health(self):
        """
        The health probe API validates the status of DNS resolution. Use this method to configure the DNS Edge
         service point behind load balancers.

        Example:

            .. code:: python

                from bluecat_libraries.service_point.api import ServicePointClient

                client = ServicePointClient(<service_point_url>)
                try:
                    client.get_v1_status_health()
                catch Exception:
                    available = False
                else:
                    available = True

        .. versionadded:: 20.6.1
        """
        response = self.http_get(self._url_for("/v1/status/health"))
        if response.status_code != 200:
            raise UnexpectedResponse(
                "Got {status_code} which is unexpected".format(status_code=response.status_code),
                response,
            )

    def get_v1_status_spdiagnostics(self):
        """
        Use this method for troubleshoot a service point. It returns the overall health status of the service point,
        the service point ID, each service's status and service version, configured forwarder IPs,
        the current local time and time zone, and policy details.

        Example:

            .. code:: python

                from bluecat_libraries.service_point.api import ServicePointClient

                client = ServicePointClient(<service_point_url>)
                status_spdiagnostics = client.get_v1_status_spdiagnostics()

        .. versionadded:: 20.6.1
        """

        response = self.http_get(self._url_for("/v1/status/spDiagnostics"))
        return response.json()



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Modules for working with BlueCat Failover API."""
from bluecat_libraries.address_manager.failover_api.client import Client

__all__ = ["Client"]



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the access right API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class AccessRightValues(StrEnum):
    """Values for types of access right in BlueCat Address Manager."""

    AddAccess = "ADD"
    ChangeAccess = "CHANGE"
    FullAccess = "FULL"
    HideAccess = "HIDE"
    ViewAccess = "VIEW"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in IP Address related API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class IPAddressState(StrEnum):
    """Values for types of IP address' states in BlueCat Address Manager."""

    UNALLOCATED = "UNALLOCATED"
    STATIC = "STATIC"
    DHCP_ALLOCATED = "DHCP_ALLOCATED"
    DHCP_FREE = "DHCP_FREE"
    DHCP_RESERVED = "DHCP_RESERVED"
    DHCP_LEASED = "DHCP_LEASED"
    RESERVED = "RESERVED"
    GATEWAY = "GATEWAY"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the DHCP custom option API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class DHCPCustomOptionType(StrEnum):
    """DHCP Custom Option Type"""

    BINARY = "BINARY"
    BOOLEAN = "BOOLEAN"
    ENCAPSULATED = "ENCAPSULATED"
    IP4 = "IP4"
    IP4_BLOCK = "IP4_BLOCK"
    IP4_MASK = "IP4_MASK"
    IP4_RANGE = "IP4_RANGE"
    SIGNED_INT_8 = "SIGNED_INT_8"
    SIGNED_INT_16 = "SIGNED_INT_16"
    SIGNED_INT_32 = "SIGNED_INT_32"
    STRING = "STRING"
    TEXT = "TEXT"
    UNSIGNED_INT_8 = "UNSIGNED_INT_8"
    UNSIGNED_INT_16 = "UNSIGNED_INT_16"
    UNSIGNED_INT_32 = "UNSIGNED_INT_32"
    UNSIGNED_INT_64 = "UNSIGNED_INT_64"



# Copyright 2020-2022 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Creating the Client object that is the interface for doing API calls.
"""
import logging
import re
import threading
from typing import Any, Optional

from requests import Session
from urllib3.util import parse_url

from bluecat_libraries.http_client import GeneralError, ErrorResponse
from bluecat_libraries.address_manager.api.rest.auto import _wadl_parser
from bluecat_libraries.address_manager.api.serialization import deserialize_joined_key_value_pairs
from bluecat_libraries.address_manager.api._version import Version as _Version

__all__ = [
    "Client",
]

_LOGGER = logging.getLogger(__name__)
_LOGIN_RESPONSE_WITH_TOKEN = re.compile(
    r"Session Token-> (?P<token>.+) <- for User : (?P<username>.+)$",
    # r'^Session Token-> (?P<token>.+) <- for User : (?P<username>.+)$',
    re.UNICODE,
)

_lock = threading.Lock()

_SERVICE_CACHE = {}


def _get_rest_service_base_url(url: str, use_https: bool = False) -> str:
    """
    Takes an URL to a BAM server and returns the base URL to the BAM's REST
    service.

    :param url: A URL to a BlueCat Address Manager. The path, query, and
        fragments portions will be ignored, even if provided
    :type url: str
    :param use_https: Whether the returned URL should be for HTTPS. Defaults
        to `False`.
    :type use_https: bool, Optional

    :return: The URL to the base path for the REST service of BAM.
    :rtype: str
    """
    parts = parse_url(url)
    scheme = "https" if use_https else (parts.scheme or "https")
    if parts.port:
        host_incl_port = "{}:{}".format(parts.host, parts.port)
    else:
        host_incl_port = parts.host
    if not host_incl_port:
        raise GeneralError("Invalid BAM API URL: {url} provided!".format(url=url))
    address = "{scheme}://{host_incl_port}/Services/REST/".format(
        scheme=scheme, host_incl_port=host_incl_port
    )
    return address


def _get_rest_service_wadl_url(rest_service_base_url: str) -> str:
    """
    Return the URL to the WADL of BAM's REST service.

    :param rest_service_base_url: Base URL of BAM's REST service.
    :type rest_service_base_url: str
    :return: The URL to the WADL of BAM's REST service.
    :rtype: str
    """
    return rest_service_base_url + "application.wadl"


def _get_rest_endpoint_base_url(rest_service_base_url: str) -> str:
    """
    Return the base URL of the endpoints of BAM's REST service.

    :param rest_service_base_url: Base URL of BAM's REST service.
    :type rest_service_base_url: str
    :return: The base URL of the endpoints of BAM's REST service.
    :rtype: str
    """
    return (
        rest_service_base_url + "v1"
    )  # TODO: Shouldn't this end in a forward slash?  # pylint: disable=fixme


def _get_service_for_url(rest_service_base_url, session):
    """
    Returns an object with automatically generated methods for the endpoints of the BAM REST API
    service.

    The returned object is cached for the given URL. Subsequent calls with the same URL will return
    the cached object. If there is no cached object satisfies the call, the service description
    of the BAM REST API is obtained and an object is created.

    :param rest_service_base_url: Base URL of the REST service of BAM.
    :type rest_service_base_url: str
    :param session: A session object to be used, if any data needs to be obtained from BAM through
        HTTP requests.
    :type session: requests.Session
    :return: Service
    """
    global _SERVICE_CACHE
    wadl_url = _get_rest_service_wadl_url(rest_service_base_url)
    if wadl_url not in _SERVICE_CACHE:
        try:
            _lock.acquire()

            # A second check, because in the time between the first check and
            # the locking another thread might have already obtained and cached
            # the necessary service.
            if wadl_url in _SERVICE_CACHE:
                return _SERVICE_CACHE[wadl_url]

            wadl = _get_rest_service_wadl(wadl_url, session)
            rest_endpoint_base_url = _get_rest_endpoint_base_url(rest_service_base_url)
            parser = _wadl_parser.WADLParser(rest_endpoint_base_url, wadl)
            _SERVICE_CACHE[wadl_url] = parser.generate_service()
        finally:
            _lock.release()
    return _SERVICE_CACHE[wadl_url]


def _get_rest_service_wadl(wadl_url: str, session: Session, *, logger=None) -> str:
    """
    Returns the WADL specification of a service as text.

    :param wadl_url: URL of the WADL specification of the service.
    :type wadl_url: str
    :param session: A session object to be used to issue HTTP requests to obtain the WADL.
    :type session: requests.Session
    :param logger: Logger for logging messages.
    :type logger: logging.Logger
    :return: A service's WADL as text.
    :rtype: str
    """
    logger = logger if logger else _LOGGER
    try:
        # NOTE: The following line should work. The expectation is that the value of `verify` of
        # the session will be honoured inside the `.get()` call. However, in the test pipeline /
        # environment that does not happen. Until test environment is updated be explicit.
        #
        # response = session.get(wadl_url)
        #
        # NOTE: Passing `verify` explicitly in the method call as a temporary workaround.
        response = session.get(wadl_url, verify=session.verify)
    except Exception as exc:
        logger.exception("Error obtaining service specification from BAM: {}".format(repr(exc)))
        raise GeneralError(str(exc)) from exc
    if not response.ok:
        message = "Error obtaining service specification from BAM: HTTP {}: {}".format(
            response.status_code, response.reason
        )
        logger.exception(message)
        raise GeneralError(message)
    return response.text


class Client:
    """Client object for handling REST requests."""

    def __init__(self, url: str, *, verify: Any = True) -> None:
        """
        ``verify`` has the same meaning as arguments of the same name commonly found in
        ``requests`` functions.
        """
        self.session = Session()
        self.session.verify = verify
        rest_service_base_url = _get_rest_service_base_url(url, use_https=url.startswith("https"))
        self._service = _get_service_for_url(rest_service_base_url, self.session)

    def close(self):
        """Release any allocated resources, e.g., an internal session."""
        self.session.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @property
    def version(self) -> Optional[str]:
        """
        Version of the BlueCat Address Manager the used service is for. Return value
        ``None`` means this client has never been logged into.

        :return: Version of BlueCat Address Manager.
        :rtype: Optional[str]
        """
        v = self._service.version
        return str(v) if v else None

    @property
    def is_authenticated(self) -> bool:
        """
        Determine whether the authentication necessary to communicate with BlueCat Address
        Manager is set.
        """
        return bool(self.session.headers["Authorization"])

    def login(self, username, password):
        """
        Wrapper for the REST login API/endpoint.

        :param username: Username of the user that logs in.
        :type username: str
        :param password: Password of the user that logs in.
        :type password: str
        :return: Access token for BlueCat Address Manager.
        :rtype: str
        """
        try:
            content = self._service.login(self.session, username, password)
        except Exception as exc:
            if password is not None and password in str(exc):
                raise GeneralError("Connection Error!") from exc
            raise

        match = _LOGIN_RESPONSE_WITH_TOKEN.search(content)
        if not match:
            raise GeneralError(
                "Invalid username or password."
            )  # Keeping the original exception message.
            # raise GeneralError('Invalid response from BAM.')

        token = match.group("token")
        # TODO: Could usernames in BAM be case-insensitive?  # pylint: disable=fixme
        # username_ = match.group('username')
        # if username != username_:
        #     raise GeneralError('Received response for different user.')

        self.set_token(token)
        return token

    def loginWithOptions(self, username, password, options):
        """
        Wrapper for the REST loginWithOptions API/endpoint.

        :param username: Username of the user that logs in.
        :type username: str
        :param password: Password of the user that logs in.
        :type password: str
        :param options: Login options.
        :type options: str
        :return: Access token for BlueCat Address Manager.
        :rtype: str
        """
        try:
            content = self._service.loginWithOptions(self.session, username, password, options)
        except Exception as e:
            if password is not None and password in str(e):
                raise GeneralError("Connection Error!") from e
            raise

        match = _LOGIN_RESPONSE_WITH_TOKEN.search(content)
        if not match:
            raise GeneralError("Invalid username or password.")

        token = match.group("token")
        self.set_token(token)
        return token

    def set_token(self, token):
        """
        Set token used for BAM authentication.

        :param token: Token to be used for BAM authentication.
        """
        self.session.headers.update({"Authorization": token})
        if not self._service.version:
            self._set_version()

    def clear_token(self):
        """
        Clear the token used for BAM authentication.
        """
        del self.session.headers["Authorization"]

    def _set_version(self) -> None:
        """Client must be authenticated in order to access BAM version information."""
        try:
            sysinfo = self.getSystemInfo()
        except ErrorResponse as e:
            if "API method cannot be invoked from a read only session" in e.message:
                _LOGGER.warning(
                    'Unable to resolve system version for "%s". Cannot check if methods satisfy minimum version requirements.',
                    self._service.url,
                )
                return
            raise
        sysinfo = deserialize_joined_key_value_pairs(sysinfo)
        version = sysinfo["version"]  # e.g. '1.2.3-456.GA.bcn'
        version = version.split("-")[0]
        self._service.version = _Version(version)

    def __getattr__(self, item):
        """
        Wraps generic REST methods (attached to a ServiceBase instance) and
        passes to them the `requests` session.

        :param item: The method attribute being called.
        :return: The result of the call.
        """
        wrapped = getattr(self._service, item)

        def func_wrapper(*args, **kwargs):
            """
            Wrapper function that passes the `requests` session to use for the
            REST call.

            :param args: Positional arguments for the REST call.
            :param kwargs: Keyword arguments for the REST call.
            :return: Result of the REST call.
            """
            return wrapped(self.session, *args, **kwargs)

        return func_wrapper



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Tools for processing a BlueCat Address Manager REST API service specification
and generating functions for calling discovered endpoints.
"""
import re
from typing import Optional
from xml.etree import ElementTree

from bluecat_libraries.address_manager.api._version import Version
from bluecat_libraries.http_client import ClientError, ErrorResponse


class ServiceBase:
    """
    A base (and empty) service representation that can be used as subject of
    extension with methods for calling Address Manager endpoints.
    The instance(s) of this class are not expected to be used as a
    client directly, but as holders of the generated wrappers.

    :param url: Base URL for the target service's endpoints. Defaults to `None`.
    :type url: str, optional
    :param version: Version of the target service. Defaults to `None`.
    :type version: str, optional
    """

    def __init__(self, url=None, version: Optional[Version] = None) -> None:
        self.url = url
        self.version = version

    def __getattr__(self, item):
        """
        User tries to make a REST call that wasn't generated. Probably means it
        doesn't exist.

        :param item: The method attribute being called.
        :raise: ClientError
        """
        raise ClientError("REST API method not found: {}".format(item))


def get_element_namespace(element):
    """
    Gets the namespace of an element.

    :param element: Element to get the namespace of.
    :return: Namespace of given element.
    """
    result = re.match("{.+}", element.tag)
    return result.group(0) if result else ""


def process_rest_response(response):
    """
    Processes response from Address Manager REST API.

    :param response:
    :return: object
    """
    if not response.ok:
        message = response.text
        if not message:
            message = "Response with HTTP status code {}".format(response.status_code)
        raise ErrorResponse(message, response)
    try:
        return response.json()
    except (AttributeError, ValueError):
        return response.text


class WADLParser:
    """
    Generator of a service for calling endpoints defined in a WADL.
    """

    def __init__(self, url, wadl, namespace=None):
        if isinstance(wadl, str):
            wadl = ElementTree.fromstring(wadl)
        if not namespace:
            namespace = get_element_namespace(wadl)

        self.endpoint_base_url = url
        self.wadl = wadl
        self.xpath_method = ".//{0}method".format(namespace)
        self.xpath_request_param = "./{0}request/{0}param".format(namespace)
        self.xpath_request_representation = "./{0}request/{0}representation".format(namespace)

    def generate_service(self):
        """
        Process the provided WADL and construct an object with generated methods for the described
        endpoints.

        :return: An object with generated methods for the endpoints of the service described in the
            provided WADL.
        :rtype: ServiceBase
        """
        target = ServiceBase(self.endpoint_base_url)
        self.generate_all_methods(target)
        return target

    def generate_all_methods(self, target):
        """
        Iterates through all REST methods defined in the WADL file and attaches
        the generated Python methods to the target instance.

        :param target: Service object to attach the generated methods to.
        :type target: object
        """
        rest_methods = self.wadl.findall(self.xpath_method)
        for method in rest_methods:
            self.generate_method(target, method)

    def generate_method(self, target, rest_method):
        """
        Generate Python method based on a WADL specification and attach it to
        the target instance.

        :param target: Service object to attach a generated method to.
        :type target: object
        :param rest_method: XML element from the WADL file, describing a REST
            method.
        """
        rest_params = rest_method.findall(self.xpath_request_param)
        rest_call_type = rest_method.attrib["name"].lower()
        rest_method_name = rest_method.attrib["id"]
        representation_param = rest_method.findall(self.xpath_request_representation)
        extra_data_type = ""
        if representation_param:
            extra_data_type = representation_param[0].attrib["mediaType"]

        endpoint_url = "{url}/{rest_call}".format(
            url=self.endpoint_base_url, rest_call=rest_method_name
        )

        def generic_rest_call(session, *args, **kwargs):
            """
            Generates a REST API call to BlueCat Address Manager based on the item the user specified.

            :param session: requests session used to make the REST call with.
            :param args: positional arguments for the REST call.
            :param kwargs: keyword arguments for the REST call.
            :return: Response for the REST call.
            """
            parameters = {}
            index = -1
            extra_arg = None
            try:
                for index, param in enumerate(rest_params):
                    parameters[param.attrib["name"]] = args[index]
            except IndexError:
                pass
            finally:
                if len(parameters) < len(rest_params) + len(representation_param):
                    for key, value in kwargs.items():
                        if any(
                            rest_params[i].attrib["name"] == key for i in range(0, len(rest_params))
                        ):
                            parameters[key] = value
                        else:
                            extra_arg = {key: value}
            http_method = getattr(session, rest_call_type)

            # NOTE: The expectation that the value of `verify` of the session will be honoured
            # inside the HTTP methods below is not satisfied. In the test pipeline / environment
            # that does not happen. Until test environment is updated we have to be explicit.
            # NOTE: Passing `verify` explicitly in the method calls as a temporary workaround.

            # Checks if you need to send json over with the request
            if extra_data_type in ["application/json", "multipart/form-data"] and len(args) + len(
                kwargs
            ) > len(rest_params):
                if extra_data_type == "multipart/form-data":
                    if not extra_arg:
                        extra_arg = {"data": args[index + 1]}
                    response = http_method(
                        endpoint_url, params=parameters, files=extra_arg, verify=session.verify
                    )
                else:
                    if not extra_arg:
                        extra_arg = args[index + 1]
                    response = http_method(
                        endpoint_url, params=parameters, json=extra_arg, verify=session.verify
                    )
            else:
                response = http_method(endpoint_url, params=parameters, verify=session.verify)
            return process_rest_response(response)

        setattr(target, rest_method.attrib["id"], generic_rest_call)
        return generic_rest_call



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# pylint: disable=C0112,C0115,C0116,W0511,W0613;
"""Functions related to the conversion of data to and from BlueCat Address Manager's API."""
import re
from typing import Union

from bluecat_libraries.http_client.exceptions import GeneralError


def deserialize_joined_key_value_pairs(value: str, item_sep="|", keyvalue_sep="=") -> dict:
    """Convert a single-string representation into a dictionary."""

    # if not value.endswith('|'):
    #     raise GeneralError('Serialized properties (key-value pairs) should end with a pipe character.')
    if value.endswith(item_sep):
        value = value[:-1]

    # NOTE: A straightforward "split by pipes and then split by equal signs" would cause erroneous interpretations,
    # because of the presence of escaped pipes in both names and values.
    # NOTE: Currently, equal signs are not escaped, although they may be present in both the name and the value of a
    # property. We can have "=====" as a key-value pair and no way of knowing is it "=" : "===", "==": "==", or
    # "===": "=". Thus, the first occurring equal sign is treated as a separator.
    pairs = []
    escaped = False
    token = []
    for char in value:
        if char == "\\":
            if escaped:
                token.append("\\")
                escaped = False
            else:
                escaped = True
            continue
        if char == item_sep:
            if escaped:
                token.append(item_sep)
                escaped = False
            else:
                if not token:
                    raise GeneralError("Empty key-value pair in serialized properties.")
                pairs.append("".join(token))
                token = []
            continue
        token.append(char)
    if escaped:
        raise GeneralError("Incomplete escape sequence in serialized properties (key-value pairs).")
    if token:
        pairs.append("".join(token))
    pairs = map(lambda item: item.split(keyvalue_sep, maxsplit=1), pairs)
    return dict(pairs)


def serialize_joined_key_value_pairs(values: dict, item_sep="|", keyvalue_sep="=") -> str:
    if not values:
        return ""

    result = item_sep.join(
        "{key}{keyvalue_sep}{value}".format(
            key=escape(key), keyvalue_sep=keyvalue_sep, value=escape(str(value))
        )
        for key, value in values.items()
    )
    return result + item_sep if result else ""


def serialize_joined_values(values: list, item_sep=",") -> str:
    if not values:
        return ""
    result = item_sep.join(str(x) for x in values)
    return result


def serialize_possible_list(value: Union[str, "list[str]", "list[list[str]]"]) -> str:
    """
    If ``value`` is a list, then either all inner elements must be lists, or no element can be a
    list. Otherwise, no behavior is guaranteed.
    """
    # Simple value case
    if not isinstance(value, list):
        return str(value)

    # Empty list case
    if not value:
        return ""

    # 1D case
    if not isinstance(value[0], list):
        return ",".join(str(x) for x in value)

    # 2D case
    out = []
    for inner in value:
        x = ",".join(str(x) for x in inner)
        out.append("{" + x + "}")
    return ",".join(out)


def deserialize_possible_list(value: str) -> Union[str, "list[str]", "list[list[str]]"]:
    # We rely on the fact that individual values do not contain `{`, `}`, or `,`

    # 2D case
    if value.startswith("{"):
        return [item.group(1).split(",") for item in re.finditer(r"{([^}]*)}", value)]

    # Empty string becomes [""]
    out = value.split(",")
    return out if len(out) > 1 else out[0]


def escape(value: str) -> str:
    return value.replace("\\", "\\\\").replace("|", "\\|")


def unescape(value: str) -> str:
    return value.replace("\\|", "|").replace("\\\\", "\\")



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values for types of objects in BlueCat Address Manager."""
from ._enum import StrEnum


class ObjectType(StrEnum):
    """Values for types of objects in BlueCat Address Manager."""

    ACL = "ACL"  # Access Control List (ACL)
    ALIAS_RECORD = "AliasRecord"
    CONFIGURATION = "Configuration"
    CUSTOM_OPTION_DEF = "CustomOptionDef"
    DENY_MAC_POOL = "DenyMACPool"
    DEPLOYMENT_ROLE = "DNSDeploymentRole"
    DEPLOYMENT_SCHEDULER = "DeploymentScheduler"
    DEVICE = "Device"
    DEVICE_SUBTYPE = "DeviceSubtype"
    DEVICE_TYPE = "DeviceType"
    DHCP_DEPLOYMENT_ROLE = "DHCPDeploymentRole"
    DHCP_MATCH_CLASS = "DHCPMatchClass"
    DHCP_RAW_OPTION = "DHCP_RAW"
    DHCP_SERVICE_OPTION = "DHCPServiceOption"
    DHCP_SUBCLASS = "DHCPSubClass"
    DHCP_VENDOR_CLIENT = "DHCPVendorClient"
    DHCP4_CLIENT_OPTION = "DHCPV4ClientOption"
    DHCP4_RANGE = "DHCP4Range"
    DHCP6_CLIENT_OPTION = "DHCPV6ClientOption"
    DHCP6_RANGE = "DHCP6Range"
    DHCP6_RAW_OPTION = "DHCPV6_RAW"  # DHCPV6 Raw Option
    DHCP6_SERVICE_OPTION = "DHCPV6ServiceOption"
    DNS = "DNS"
    DNS_OPTION = "DNSOption"
    DNS_RAW_OPTION = "DNS_RAW"  # DNS Raw Option
    DNSSEC_SIGNING_POLICY = "DNSSECSigningPolicy"
    ENTITY = "Entity"  # Generic type
    ENUM_NUMBER = "EnumNumber"
    ENUM_ZONE = "EnumZone"
    EXTERNAL_HOST_RECORD = "ExternalHostRecord"
    GENERIC_RECORD = "GenericRecord"
    HINFO_RECORD = "HINFORecord"
    HOST_RECORD = "HostRecord"
    INTERFACE_ID = "InterfaceID"
    INTERNAL_ROOT_ZONE = "InternalRootZone"
    IP4_ADDRESS = "IP4Address"
    IP4_BLOCK = "IP4Block"
    IP4_IP_GROUP = "IP4IPGroup"
    IP4_NETWORK = "IP4Network"
    IP4_NETWORK_TEMPLATE = "IP4NetworkTemplate"
    IP4_RECONCILIATION_POLICY = "IP4ReconciliationPolicy"
    IP6_ADDRESS = "IP6Address"
    IP6_BLOCK = "IP6Block"
    IP6_NETWORK = "IP6Network"
    KERBEROS = "Kerberos"
    KERBEROS_REALM = "KerberosRealm"
    LDAP = "LDAP"
    LOCATION = "Location"
    MAC_ADDRESS = "MACAddress"
    MAC_POOL = "MACPool"
    MX_RECORD = "MXRecord"
    NAPTR_RECORD = "NAPTRRecord"
    NETWORK_INTERFACE = "NetworkInterface"
    NETWORK_SERVER_INTERFACE = "NetworkServerInterface"
    PUBLISHED_SERVER_INTERFACE = "PublishedServerInterface"
    RADIUS = "Radius"
    RECORD_WITH_LINK = "RecordWithLink"
    RESPONSE_POLICY = "ResponsePolicy"
    RP_ZONE = "RPZone"
    SERVER = "Server"
    SRV_RECORD = "SRVRecord"
    START_OF_AUTHORITY = "StartOfAuthority"
    TAG = "Tag"
    TAG_GROUP = "TagGroup"
    TFTP_DEPLOYMENT_ROLE = "TFTPDeploymentRole"
    TFTP_FILE = "TFTPFile"
    TFTP_FOLDER = "TFTPFolder"
    TFTP_GROUP = "TFTPGroup"
    TSIG_KEY = "TSIGKey"
    TXT_RECORD = "TXTRecord"
    USER = "User"
    USER_GROUP = "UserGroup"
    VENDOR_CLIENT_OPTION = "VendorClientOption"
    VENDOR_PROFILE = "VendorProfile"
    VENDOR_PROFILE_DEF = "VendorOptionDef"
    VIEW = "View"
    VIRTUAL_INTERFACE = "VirtualInterface"
    ZONE = "Zone"
    ZONE_TEMPLATE = "ZoneTemplate"
